import numpy as np
import h5py
from scipy.interpolate import interp1d
import os
import math
import warnings
from quanestimation import QJL
import quanestimation.MeasurementOpt as Measure
from quanestimation.Common.Common import gramschmidt, sic_povm


class MeasurementSystem:
    """
    Attributes
    ----------
    > **mtype:** `string`
        -- The type of scenarios for the measurement optimization. Options are:  
        "projection" (default) -- Optimization of rank-one projective measurements.  
        "input" -- Find the optimal linear combination or the optimal rotated measurement 
        of a given set of POVM.

    > **minput:** `list`
        -- In the case of optimization of rank-one projective measurements, the 
        `minput` should keep empty. For finding the optimal linear combination and 
        the optimal rotated measurement of a given set of POVM, the input rule are 
        `minput=["LC", [Pi1,Pi2,...], m]` and `minput=["LC", [Pi1,Pi2,...]]` respectively.
        Here `[Pi1,Pi2,...]` represents a list of input POVM and `m` is the number of operators 
        of the output measurement. 

    > **savefile:** `bool`
        -- Whether or not to save all the measurements.  
        If set `True` then the measurements and the values of the 
        objective function obtained in all episodes will be saved during 
        the training. If set `False` the measurement in the final 
        episode and the values of the objective function in all episodes 
        will be saved.

   > **measurement0:** `list of arrays`
        -- Initial guesses of measurements.

    > **seed:** `int`
        -- Random seed.

    > **eps:** `float`
        -- Machine epsilon.

    > **load:** `bool`
        -- Whether or not to load measurements in the current location.  
        If set `True` then the program will load measurement from "measurements.csv"
        file in the current location and use it as the initial measurement.

    > **dyn_method:** `string`
        -- The method for solving the Lindblad dynamcs. Options are:
        "expm" (default) -- matrix exponential.
        "ode" -- ordinary differential equation solvers.  
    """

    def __init__(self, mtype, minput, savefile, measurement0, seed, eps, load):

        self.mtype = mtype
        self.minput = minput
        self.savefile = savefile
        self.eps = eps
        self.seed = seed
        self.load = load
        self.measurement0 = measurement0

    def load_save(self, mnum, max_episode):
        if os.path.exists("measurements.dat"):
            fl = h5py.File("measurements.dat",'r')
            dset = fl["measurements"]
            if self.savefile:
                mdata = np.array([[np.array(fl[fl[dset[i]][j]]).view('complex') for j in range(mnum)] for i in range(max_episode)])
            else:
                mdata = np.array([np.array(fl[dset[j]]).view('complex') for j in range(mnum)])
            np.save("measurements", mdata)
        else: pass

    def dynamics(self, tspan, rho0, H0, dH, Hc=[], ctrl=[], decay=[], dyn_method="expm"):
        r"""
        The dynamics of a density matrix is of the form  
        
        \begin{align}
        \partial_t\rho &=\mathcal{L}\rho \nonumber \\
        &=-i[H,\rho]+\sum_i \gamma_i\left(\Gamma_i\rho\Gamma^{\dagger}_i-\frac{1}{2}
        \left\{\rho,\Gamma^{\dagger}_i \Gamma_i \right\}\right),
        \end{align} 

        where $\rho$ is the evolved density matrix, H is the Hamiltonian of the 
        system, $\Gamma_i$ and $\gamma_i$ are the $i\mathrm{th}$ decay 
        operator and corresponding decay rate.

        Parameters
        ----------
        > **tspan:** `array`
            -- Time length for the evolution.

        > **rho0:** `matrix`
            -- Initial state (density matrix).

        > **H0:** `matrix or list`
            -- Free Hamiltonian. It is a matrix when the free Hamiltonian is time-
            independent and a list of length equal to `tspan` when it is time-dependent.

        > **dH:** `list`
            -- Derivatives of the free Hamiltonian on the unknown parameters to be 
            estimated. For example, dH[0] is the derivative vector on the first 
            parameter.

        > **Hc:** `list`
            -- Control Hamiltonians.

        > **ctrl:** `list of arrays`
            -- Control coefficients.

        > **decay:** `list`
            -- Decay operators and the corresponding decay rates. Its input rule is 
            decay=[[$\Gamma_1$, $\gamma_1$], [$\Gamma_2$,$\gamma_2$],...], where $\Gamma_1$ 
            $(\Gamma_2)$ represents the decay operator and $\gamma_1$ $(\gamma_2)$ is the 
            corresponding decay rate.

        > **dyn_method:** `string`
            -- Setting the method for solving the Lindblad dynamics. Options are:  
            "expm" (default) -- Matrix exponential.  
            "ode" -- Solving the differential equations directly.
        """

        self.tspan = tspan
        self.rho0 = np.array(rho0, dtype=np.complex128)

        self.dynamics_type = "dynamics"

        if len(dH) == 1:
            self.para_type = "single_para"
        else:
            self.para_type = "multi_para"

        if dyn_method == "expm":
            self.dyn_method = "Expm"
        elif dyn_method == "ode":
            self.dyn_method = "Ode"

        if self.mtype == "projection":
            self.M_num = len(self.rho0)
            QJLType_C = QJL.Vector[QJL.Vector[QJL.ComplexF64]]
            
            if self.measurement0 == []:
                np.random.seed(self.seed)
                M = [[] for i in range(len(self.rho0))]
                for i in range(len(self.rho0)):
                    r_ini = 2 * np.random.random(len(self.rho0)) - np.ones(
                        len(self.rho0)
                    )
                    r = r_ini / np.linalg.norm(r_ini)
                    phi = 2 * np.pi * np.random.random(len(self.rho0))
                    M[i] = [r[j] * np.exp(1.0j * phi[j]) for j in range(len(self.rho0))]
                self.C = QJL.convert(QJLType_C, gramschmidt(np.array(M)))
                self.measurement0 = QJL.Vector([self.C])
            else:
                self.C = [self.measurement0[0][i] for i in range(len(self.rho0))]
                self.C = QJL.convert(QJLType_C, self.C)
                self.measurement0 = QJL.Vector([self.C])
            self.opt = QJL.Mopt_Projection(M=self.C, seed=self.seed)

        elif self.mtype == "input":
            if self.minput[0] == "LC":
                self.M_num = self.minput[2]
                ## optimize the combination of a set of SIC-POVM
                if self.minput[1] == []:
                    file_path = os.path.join(
                        os.path.dirname(os.path.dirname(__file__)),
                        "sic_fiducial_vectors/d%d.txt" % (len(self.rho0)),
                    )
                    data = np.loadtxt(file_path)
                    fiducial = data[:, 0] + data[:, 1] * 1.0j
                    fiducial = np.array(fiducial).reshape(len(fiducial), 1)
                    self.povm_basis = sic_povm(fiducial)
                else:
                    ## optimize the combination of a set of given POVMs
                    if type(self.minput[1]) != list:
                        raise TypeError("The given POVMs should be a list!")
                    else:
                        accu = len(str(int(1 / self.eps))) - 1
                        for i in range(len(self.minput[1])):
                            val, vec = np.linalg.eig(self.minput[1])
                            if np.all(val.round(accu) >= 0):
                                pass
                            else:
                                raise TypeError(
                                    "The given POVMs should be semidefinite!"
                                )
                        M = np.zeros(
                            (len(self.rho0), len(self.rho0)), dtype=np.complex128
                        )
                        for i in range(len(self.minput[1])):
                            M += self.minput[1][i]
                        if np.all(M.round(accu) - np.identity(len(self.rho0)) == 0):
                            pass
                        else:
                            raise TypeError(
                                "The sum of the given POVMs should be identity matrix!"
                            )
                        self.povm_basis = [
                            np.array(x, dtype=np.complex128) for x in self.minput[1]
                        ]

                if self.measurement0 == []:
                    np.random.seed(self.seed)
                    self.B = [
                        np.random.random(len(self.povm_basis))
                        for i in range(self.M_num)
                    ]
                    self.measurement0 = [self.B]
                elif len(self.measurement0) >= 1:
                    self.B = [self.measurement0[0][i] for i in range(self.M_num)]
                    self.measurement0 = [[m for m in m0] for m0 in self.measurement0]
                    
                
                QJLType_B = QJL.Vector[QJL.Vector[QJL.Float64]]
                QJLType_pb = QJL.Vector[QJL.Matrix[QJL.ComplexF64]]
                QJLType_m0 = QJL.Vector[QJL.Vector[QJL.Vector[QJL.ComplexF64]]]
                self.B = QJL.convert(QJLType_B, self.B)
                self.povm_basis = QJL.convert(QJLType_pb, self.povm_basis)
                self.measurement0 = QJL.convert(QJLType_m0, self.measurement0)
                
                self.opt = QJL.Mopt_LinearComb(
                    B=self.B, POVM_basis=self.povm_basis, M_num=self.M_num, seed=self.seed
                )

            elif self.minput[0] == "rotation":
                self.M_num = len(self.minput[1])
                ## optimize the coefficients of the rotation matrix
                if type(self.minput[1]) != list:
                    raise TypeError("The given POVMs should be a list!")
                else:
                    if self.minput[1] == []:
                        raise TypeError("The initial POVM should not be empty!")
                    accu = len(str(int(1 / self.eps))) - 1
                    for i in range(len(self.minput[1])):
                        val, vec = np.linalg.eig(self.minput[1])
                        if np.all(val.round(accu) >= 0):
                            pass
                        else:
                            raise TypeError("The given POVMs should be semidefinite!")
                    M = np.zeros((len(self.rho0), len(self.rho0)), dtype=np.complex128)
                    for i in range(len(self.minput[1])):
                        M += self.minput[1][i]
                    if np.all(M.round(accu) - np.identity(len(self.rho0)) == 0):
                        pass
                    else:
                        raise TypeError(
                            "The sum of the given POVMs should be identity matrix!"
                        )
                    self.povm_basis = [
                        np.array(x, dtype=np.complex128) for x in self.minput[1]
                    ]
                    self.mtype = "rotation"

                if self.measurement0 == []:
                    np.random.seed(self.seed)
                    self.s = np.random.random(len(self.rho0) ** 2)
                    self.measurement0 = [self.s]
                elif len(self.measurement0) >= 1:
                    self.s = [
                        self.measurement0[0][i]
                        for i in range(len(self.rho0) * len(self.rho0))
                    ]

                self.s = QJL.Vector(self.s)
                QJLType_pb = QJL.Vector[QJL.Matrix[QJL.ComplexF64]]
                self.povm_basis = QJL.convert(QJLType_pb, self.povm_basis)
                self.opt = QJL.Mopt_Rotation(
                    s=self.s, POVM_basis=self.povm_basis, Lambda=[], seed=self.seed
                )

            else:
                raise ValueError(
                    "{!r} is not a valid value for the first input of minput, supported values are 'LC' and 'rotation'.".format(
                        self.minput[0]
                    )
                )
        else:
            raise ValueError(
                "{!r} is not a valid value for mtype, supported values are 'projection' and 'input'.".format(
                    self.mtype
                )
            )

        if Hc == [] or ctrl == []:
            if type(H0) == np.ndarray:
                self.freeHamiltonian = np.array(H0, dtype=np.complex128)
            else:
                self.freeHamiltonian = [np.array(x, dtype=np.complex128) for x in H0]
        else:
            ctrl_num = len(ctrl)
            Hc_num = len(Hc)
            if Hc_num < ctrl_num:
                raise TypeError(
                    "There are %d control Hamiltonians but %d coefficients sequences: too many coefficients sequences"
                    % (Hc_num, ctrl_num)
                )
            elif Hc_num > ctrl_num:
                warnings.warn(
                    "Not enough coefficients sequences: there are %d control Hamiltonians but %d coefficients sequences. The rest of the control sequences are set to be 0."
                    % (Hc_num, ctrl_num),
                    DeprecationWarning,
                )
                for i in range(Hc_num - ctrl_num):
                    ctrl = np.concatenate((ctrl, np.zeros(len(ctrl[0]))))
            else: pass

            if len(ctrl[0]) == 1:
                if type(H0) == np.ndarray:
                    H0 = np.array(H0, dtype=np.complex128)
                    Hc = [np.array(x, dtype=np.complex128) for x in Hc]
                    Htot = H0 + sum([Hc[i] * ctrl[i][0] for i in range(ctrl_num)])
                    self.freeHamiltonian = np.array(Htot, dtype=np.complex128)
                else:
                    H0 = [np.array(x, dtype=np.complex128) for x in H0]
                    Htot = []
                    for i in range(len(H0)):
                        Htot.append(
                            H0[i] + sum([Hc[i] * ctrl[i][0] for i in range(ctrl_num)])
                        )
                    self.freeHamiltonian = [
                        np.array(x, dtype=np.complex128) for x in Htot
                    ]
            else:
                if type(H0) != np.ndarray:
                    #### linear interpolation  ####
                    f = interp1d(self.tspan, H0, axis=0)
                else: pass
                number = math.ceil((len(self.tspan) - 1) / len(ctrl[0]))
                if len(self.tspan) - 1 % len(ctrl[0]) != 0:
                    tnum = number * len(ctrl[0])
                    self.tspan = np.linspace(self.tspan[0], self.tspan[-1], tnum + 1)
                    if type(H0) != np.ndarray:
                        H0_inter = f(self.tspan)
                        H0 = [np.array(x, dtype=np.complex128) for x in H0_inter]
                    else: pass
                else: pass

                if type(H0) == np.ndarray:
                    H0 = np.array(H0, dtype=np.complex128)
                    Hc = [np.array(x, dtype=np.complex128) for x in Hc]
                    ctrl = [np.array(ctrl[i]).repeat(number) for i in range(len(Hc))]
                    Htot = []
                    for i in range(len(ctrl[0])):
                        S_ctrl = sum([Hc[j] * ctrl[j][i] for j in range(len(ctrl))])
                        Htot.append(H0 + S_ctrl)
                    self.freeHamiltonian = [
                        np.array(x, dtype=np.complex128) for x in Htot
                    ]
                else:
                    H0 = [np.array(x, dtype=np.complex128) for x in H0]
                    Hc = [np.array(x, dtype=np.complex128) for x in Hc]
                    ctrl = [np.array(ctrl[i]).repeat(number) for i in range(len(Hc))]
                    Htot = []
                    for i in range(len(ctrl[0])):
                        S_ctrl = sum([Hc[j] * ctrl[j][i] for j in range(len(ctrl))])
                        Htot.append(H0[i] + S_ctrl)
                    self.freeHamiltonian = [
                        np.array(x, dtype=np.complex128) for x in Htot
                    ]

        if type(dH) != list:
            raise TypeError("The derivative of Hamiltonian should be a list!")

        if dH == []:
            dH = [np.zeros((len(self.rho0), len(self.rho0)))]
        self.Hamiltonian_derivative = [np.array(x, dtype=np.complex128) for x in dH]

        if decay == []:
            decay_opt = [np.zeros((len(self.rho0), len(self.rho0)))]
            self.gamma = [0.0]
        else:
            decay_opt = [decay[i][0] for i in range(len(decay))]
            self.gamma = [decay[i][1] for i in range(len(decay))]
        self.decay_opt = [np.array(x, dtype=np.complex128) for x in decay_opt]

        if any(self.gamma):
            self.dynamic = QJL.Lindblad(
                self.freeHamiltonian,
                self.Hamiltonian_derivative,
                self.rho0,
                self.tspan,
                self.decay_opt,
                self.gamma,
                dyn_method = self.dyn_method,
            )
        else:
            self.dynamic = QJL.Lindblad(
                self.freeHamiltonian,
                self.Hamiltonian_derivative,
                self.rho0,
                self.tspan,
                dyn_method = self.dyn_method,
            )
        self.output = QJL.Output(self.opt, save=self.savefile)
        
        self.dynamics_type = "dynamics"


    def Kraus(self, rho0, K, dK):
        r"""
        The parameterization of a state is
        \begin{align}
        \rho=\sum_i K_i\rho_0K_i^{\dagger},
        \end{align} 

        where $\rho$ is the evolved density matrix, $K_i$ is the Kraus operator.

        Parameters
        ----------
        > **rho0:** `matrix`
            -- Initial state (density matrix).

        > **K:** `list`
            -- Kraus operators.

        > **dK:** `list`
            -- Derivatives of the Kraus operators on the unknown parameters to be 
            estimated. For example, dK[0] is the derivative vector on the first 
            parameter.
        """
        k_num = len(K)
        para_num = len(dK[0])
        self.para_num = para_num
        self.dK = [
            [np.array(dK[i][j], dtype=np.complex128) for j in range(para_num)]
            for i in range(k_num)
        ]
        self.rho0 = np.array(rho0, dtype=np.complex128)
        self.K = [np.array(x, dtype=np.complex128) for x in K]

        if para_num == 1:
            self.para_type = "single_para"
        else:
            self.para_type = "multi_para"

        if self.mtype == "projection":
            self.M_num = len(self.rho0)
            if self.measurement0 == []:
                np.random.seed(self.seed)
                M = [[] for i in range(len(self.rho0))]
                for i in range(len(self.rho0)):
                    r_ini = 2 * np.random.random(len(self.rho0)) - np.ones(
                        len(self.rho0)
                    )
                    r = r_ini / np.linalg.norm(r_ini)
                    phi = 2 * np.pi * np.random.random(len(self.rho0))
                    M[i] = [r[j] * np.exp(1.0j * phi[j]) for j in range(len(self.rho0))]
                self.C = gramschmidt(np.array(M))
                self.measurement0 = [self.C]
            else:
                self.C = [self.measurement0[0][i] for i in range(len(self.rho0))]
                self.C = [np.array(x, dtype=np.complex128) for x in self.C]
            self.opt = QJL.Mopt_Projection(M=self.C, seed=self.seed)

        elif self.mtype == "input":
            if self.minput[0] == "LC":
                self.M_num = self.minput[2]
                ## optimize the combination of a set of SIC-POVM
                if self.minput[1] == []:
                    file_path = os.path.join(
                        os.path.dirname(os.path.dirname(__file__)),
                        "sic_fiducial_vectors/d%d.txt" % (len(self.rho0)),
                    )
                    data = np.loadtxt(file_path)
                    fiducial = data[:, 0] + data[:, 1] * 1.0j
                    fiducial = np.array(fiducial).reshape(len(fiducial), 1)
                    self.povm_basis = sic_povm(fiducial)
                else:
                    ## optimize the combination of a set of given POVMs
                    if type(self.minput[1]) != list:
                        raise TypeError("The given POVMs should be a list!")
                    else:
                        accu = len(str(int(1 / self.eps))) - 1
                        for i in range(len(self.minput[1])):
                            val, vec = np.linalg.eig(self.minput[1])
                            if np.all(val.round(accu) >= 0):
                                pass
                            else:
                                raise TypeError(
                                    "The given POVMs should be semidefinite!"
                                )
                        M = np.zeros(
                            (len(self.rho0), len(self.rho0)), dtype=np.complex128
                        )
                        for i in range(len(self.minput[1])):
                            M += self.minput[1][i]
                        if np.all(M.round(accu) - np.identity(len(self.rho0)) == 0):
                            pass
                        else:
                            raise TypeError(
                                "The sum of the given POVMs should be identity matrix!"
                            )
                        self.povm_basis = [
                            np.array(x, dtype=np.complex128) for x in self.minput[1]
                        ]

                if self.measurement0 == []:
                    np.random.seed(self.seed)
                    self.B = [
                        np.random.random(len(self.povm_basis))
                        for i in range(self.M_num)
                    ]
                    self.measurement0 = [np.array(self.B)]
                elif len(self.measurement0) >= 1:
                    self.B = [
                        self.measurement0[0][i] for i in range(len(self.povm_basis))
                    ]
                self.opt = QJL.Mopt_LinearComb(
                    B=self.B, POVM_basis=self.povm_basis, M_num=self.M_num, seed=self.seed
                )

            elif self.minput[0] == "rotation":
                self.M_num = len(self.minput[1])
                ## optimize the coefficients of the rotation matrix
                if type(self.minput[1]) != list:
                    raise TypeError("The given POVMs should be a list!")
                else:
                    if self.minput[1] == []:
                        raise TypeError("The initial POVM should not be empty!")
                    accu = len(str(int(1 / self.eps))) - 1
                    for i in range(len(self.minput[1])):
                        val, vec = np.linalg.eig(self.minput[1])
                        if np.all(val.round(accu) >= 0):
                            pass
                        else:
                            raise TypeError("The given POVMs should be semidefinite!")
                    M = np.zeros((len(self.rho0), len(self.rho0)), dtype=np.complex128)
                    for i in range(len(self.minput[1])):
                        M += self.minput[1][i]
                    if np.all(M.round(accu) - np.identity(len(self.rho0)) == 0):
                        pass
                    else:
                        raise TypeError(
                            "The sum of the given POVMs should be identity matrix!"
                        )
                    self.povm_basis = [
                        np.array(x, dtype=np.complex128) for x in self.minput[1]
                    ]
                    self.mtype = "rotation"

                if self.measurement0 == []:
                    np.random.seed(self.seed)
                    self.s = np.random.random(len(self.rho0) ** 2)
                    self.measurement0 = [self.s]
                elif len(self.measurement0) >= 1:
                    self.s = [
                        self.measurement0[0][i]
                        for i in range(len(self.rho0) * len(self.rho0))
                    ]

                self.opt = QJL.Mopt_Rotation(
                    s=self.s, POVM_basis=self.povm_basis, Lambda=[], seed=self.seed
                )

            else:
                raise ValueError(
                    "{!r} is not a valid value for the first input of minput, supported values are 'LC' and 'rotation'.".format(
                        self.minput[0]
                    )
                )
        else:
            raise ValueError(
                "{!r} is not a valid value for mtype, supported values are 'projection' and 'input'.".format(
                    self.mtype
                )
            )

        self.dynamic = QJL.Kraus(self.rho0, self.K, self.dK)
        self.output = QJL.Output(self.opt, save=self.savefile)

        self.dynamics_type = "Kraus"

    def CFIM(self, W=[]):
        r"""
        Choose CFI or $\mathrm{Tr}(WI^{-1})$ as the objective function. 
        In single parameter estimation the objective function is CFI and 
        in multiparameter estimation it will be $\mathrm{Tr}(WI^{-1})$.

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.
        """

        if self.dynamics_type == "dynamics":
            if W == []:
                W = np.eye(len(self.Hamiltonian_derivative))
            self.W = W
        elif self.dynamics_type == "Kraus":
            if W == []:
                W = np.eye(self.para_num)
            self.W = W
        else:
            raise ValueError(
                "Supported type of dynamics are Lindblad and Kraus."
                )

        self.obj = QJL.CFIM_obj(
            [], self.W, self.eps, self.para_type
        )  #### m=[]
        system = QJL.QuanEstSystem(
            self.opt, self.alg, self.obj, self.dynamic, self.output
        )
        QJL.run(system)
        max_num = self.max_episode if type(self.max_episode) == int else self.max_episode[0]
        self.load_save(self.M_num, max_num)


def MeasurementOpt(
    mtype="projection", minput=[], savefile=False, method="DE", **kwargs
):

    if method == "AD":
        return Measure.AD_Mopt(mtype, minput, savefile=savefile, **kwargs)
    elif method == "PSO":
        return Measure.PSO_Mopt(mtype, minput, savefile=savefile, **kwargs)
    elif method == "DE":
        return Measure.DE_Mopt(mtype, minput, savefile=savefile, **kwargs)
    else:
        raise ValueError(
            "{!r} is not a valid value for method, supported values are 'AD', 'PSO' and 'DE'.".format(
                method
            )
        )


def csv2npy_measurements(M, num):
    n = int(np.sqrt(len(M[0])))
    N = int(len(M) / num)
    M_save = []
    for mi in range(N):
        M_tp = M[mi * num : (mi + 1) * num]
        M_mi = [M_tp[i].reshape(n, n).T for i in range(num)]
        M_save.append(M_mi)
    np.save("measurements", M_save)


def load_measurements(M, num, indx=-1):
    n = int(np.sqrt(len(M[0])))
    N = int(len(M) / num)
    M_save = []
    for mi in range(N):
        M_tp = M[mi * num : (mi + 1) * num]
        M_mi = [M_tp[i].reshape(n, n).T for i in range(num)]
        M_save.append(M_mi)
    return M_save[indx]



from quanestimation import QJL
import quanestimation.MeasurementOpt.MeasurementStruct as Measurement


class PSO_Mopt(Measurement.MeasurementSystem):
    """
    Attributes
    ----------
    > **savefile:** `bool`
        -- Whether or not to save all the measurements.  
        If set `True` then the measurements and the values of the 
        objective function obtained in all episodes will be saved during 
        the training. If set `False` the measurement in the final 
        episode and the values of the objective function in all episodes 
        will be saved.

    > **p_num:** `int`
        -- The number of particles.

    > **measurement0:** `list of arrays`
        -- Initial guesses of measurements.

    > **max_episode:** `int or list`
        -- If it is an integer, for example max_episode=1000, it means the 
        program will continuously run 1000 episodes. However, if it is an
        array, for example max_episode=[1000,100], the program will run 
        1000 episodes in total but replace measurements of all  the particles 
        with global best every 100 episodes.
  
    > **c0:** `float`
        -- The damping factor that assists convergence, also known as inertia weight.

    > **c1:** `float`
        -- The exploitation weight that attracts the particle to its best previous 
        position, also known as cognitive learning factor.

    > **c2:** `float`
        -- The exploitation weight that attracts the particle to the best position  
        in the neighborhood, also known as social learning factor.

    > **seed:** `int`
        -- Random seed.

    > **eps:** `float`
        -- Machine epsilon.

    > **load:** `bool`
        -- Whether or not to load measurements in the current location.  
        If set `True` then the program will load measurement from "measurements.csv"
        file in the current location and use it as the initial measurement.
    """

    def __init__(
        self,
        mtype,
        minput,
        savefile=False,
        p_num=10,
        measurement0=[],
        max_episode=[1000, 100],
        c0=1.0,
        c1=2.0,
        c2=2.0,
        seed=1234,
        eps=1e-8,
        load=False,
    ):

        Measurement.MeasurementSystem.__init__(
            self, mtype, minput, savefile, measurement0, seed, eps, load
        )

        self.p_num = p_num
        is_int = isinstance(max_episode, int)
        self.max_episode = max_episode if is_int else QJL.Vector[QJL.Int64](max_episode)
        self.c0 = c0
        self.c1 = c1
        self.c2 = c2
        self.seed = seed

    def CFIM(self, W=[]):
        r"""
        Choose CFI or $\mathrm{Tr}(WI^{-1})$ as the objective function. 
        In single parameter estimation the objective function is CFI and 
        in multiparameter estimation it will be $\mathrm{Tr}(WI^{-1})$.

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.
        """
        ini_particle = (self.measurement0,)
        self.alg = QJL.PSO(
            self.max_episode,
            self.p_num,
            ini_particle,
            self.c0,
            self.c1,
            self.c2,
        )
        
        super().CFIM(W)



from quanestimation.MeasurementOpt.MeasurementStruct import (
    MeasurementSystem,
    MeasurementOpt,
    csv2npy_measurements,
)
from quanestimation.MeasurementOpt.AD_Mopt import (
    AD_Mopt,
)
from quanestimation.MeasurementOpt.PSO_Mopt import (
    PSO_Mopt,
)
from quanestimation.MeasurementOpt.DE_Mopt import (
    DE_Mopt,
)

__all__ = [
    "MeasurementSystem",
    "MeasurementOpt",
    "AD_Mopt",
    "PSO_Mopt",
    "DE_Mopt",
    "csv2npy_measurements",
]



from quanestimation.BayesianBound.BayesCramerRao import (
    BCFIM,
    BQFIM,
    BCRB,
    BQCRB,
    QVTB,
    VTB,
    OBB,
)
from quanestimation.BayesianBound.ZivZakai import (
    QZZB,
)
from quanestimation.BayesianBound.BayesEstimation import (
    Bayes,
    MLE,
    BCB,
    BayesCost
)

__all__ = [
    "BCFIM",
    "BQFIM",
    "BCRB",
    "BQCRB",
    "OBB",
    "QVTB",
    "VTB",
    "QZZB",
    "Bayes",
    "MLE",
    "BCB",
    "BayesCost",
]



from quanestimation import QJL
import quanestimation.ComprehensiveOpt.ComprehensiveStruct as Comp


class PSO_Compopt(Comp.ComprehensiveSystem):
    """
    Attributes
    ----------
    > **savefile:** `bool`
        -- Whether or not to save all the optimized variables (probe states, 
        control coefficients and measurements).  
        If set `True` then the optimized variables and the values of the 
        objective function obtained in all episodes will be saved during 
        the training. If set `False` the optimized variables in the final 
        episode and the values of the objective function in all episodes 
        will be saved.

    > **p_num:** `int`
        -- The number of particles.

    > **psi0:** `list of arrays`
        -- Initial guesses of states.

    > **ctrl0:** `list of arrays`
        -- Initial guesses of control coefficients.

    > **measurement0:** `list of arrays`
        -- Initial guesses of measurements.

    > **max_episode:** `int or list`
        -- If it is an integer, for example max_episode=1000, it means the 
        program will continuously run 1000 episodes. However, if it is an
        array, for example max_episode=[1000,100], the program will run 
        1000 episodes in total but replace states of all  the particles 
        with global best every 100 episodes.
  
    > **c0:** `float`
        -- The damping factor that assists convergence, also known as inertia weight.

    > **c1:** `float`
        -- The exploitation weight that attracts the particle to its best previous 
        position, also known as cognitive learning factor.

    > **c2:** `float`
        -- The exploitation weight that attracts the particle to the best position  
        in the neighborhood, also known as social learning factor.

    > **seed:** `int`
        -- Random seed.

    > **eps:** `float`
        -- Machine epsilon.
    """

    def __init__(
        self,
        savefile=False,
        p_num=10,
        psi0=[],
        ctrl0=[],
        measurement0=[],
        max_episode=[1000, 100],
        c0=1.0,
        c1=2.0,
        c2=2.0,
        seed=1234,
        eps=1e-8,
    ):

        Comp.ComprehensiveSystem.__init__(
            self, savefile, psi0, ctrl0, measurement0, seed, eps
        )

        self.p_num = p_num
        self.max_episode = max_episode
        self.c0 = c0
        self.c1 = c1
        self.c2 = c2
        self.seed = seed

    def SC(self, W=[], M=[], target="QFIM", LDtype="SLD"):
        """
        Comprehensive optimization of the probe state and control (SC).

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.

        > **M:** `list of matrices`
            -- A set of positive operator-valued measure (POVM). The default measurement 
            is a set of rank-one symmetric informationally complete POVM (SIC-POVM).

        > **target:** `string`
            -- Objective functions for searching the minimum time to reach the given 
            value of the objective function. Options are:  
            "QFIM" (default) -- choose QFI (QFIM) as the objective function.  
            "CFIM" -- choose CFI (CFIM) as the objective function.  
            "HCRB" -- choose HCRB as the objective function.  

        > **LDtype:** `string`
            -- Types of QFI (QFIM) can be set as the objective function. Options are:  
            "SLD" (default) -- QFI (QFIM) based on symmetric logarithmic derivative (SLD).  
            "RLD" -- QFI (QFIM) based on right logarithmic derivative (RLD).  
            "LLD" -- QFI (QFIM) based on left logarithmic derivative (LLD). 

        **Note:** 
            SIC-POVM is calculated by the Weyl-Heisenberg covariant SIC-POVM fiducial state 
            which can be downloaded from [here](http://www.physics.umb.edu/Research/QBism/
            solutions.html).
        """
        ini_particle = (
            self.psi, 
            self.ctrl0
       )
        
        self.alg = QJL.PSO(
            QJL.Vector[QJL.Int64](self.max_episode),
            self.p_num,
            ini_particle,
            self.c0,
            self.c1,
            self.c2,
        )
        super().SC(W, M, target, LDtype)

    def CM(self, rho0, W=[]):
        """
        Comprehensive optimization of the control and measurement (CM).

        Parameters
        ----------
        > **rho0:** `matrix`
            -- Initial state (density matrix).

        > **W:** `matrix`
            -- Weight matrix.
        """
        ini_particle = (self.ctrl0, self.measurement0)
        self.alg = QJL.PSO(
            QJL.Vector[QJL.Int64](self.max_episode),
            self.p_num,
            ini_particle,
            self.c0,
            self.c1,
            self.c2,
        )

        super().CM(rho0, W)

    def SM(self, W=[]):
        """
        Comprehensive optimization of the probe state and measurement (SM).

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.
        """
        ini_particle = (self.psi, self.measurement0)
        self.alg = QJL.PSO(
            QJL.Vector[QJL.Int64](self.max_episode),
            self.p_num,
            ini_particle,
            self.c0,
            self.c1,
            self.c2,
        )

        super().SM(W)

    def SCM(self, W=[]):
        """
        Comprehensive optimization of the probe state, the control and measurements (SCM).

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.
        """
        ini_particle = (self.psi, self.ctrl0, self.measurement0)
        self.alg = QJL.PSO(
            QJL.Vector[QJL.Int64](self.max_episode),
            self.p_num,
            ini_particle,
            self.c0,
            self.c1,
            self.c2,
        )
        
        super().SCM(W)



from quanestimation import QJL
import quanestimation.ComprehensiveOpt.ComprehensiveStruct as Comp


class AD_Compopt(Comp.ComprehensiveSystem):
    """
    Attributes
    ----------
    > **savefile:** `bool`
        -- Whether or not to save all the optimized variables (probe states, 
        control coefficients and measurements).  
        If set `True` then the optimized variables and the values of the 
        objective function obtained in all episodes will be saved during 
        the training. If set `False` the optimized variables in the final 
        episode and the values of the objective function in all episodes 
        will be saved.

    > **Adam:** `bool`
        -- Whether or not to use Adam for updating.

    > **psi0:** `list of arrays`
        -- Initial guesses of states.

    > **ctrl0:** `list of arrays`
        -- Initial guesses of control coefficients.

    > **measurement0:** `list of arrays`
        -- Initial guesses of measurements.

    > **max_episode:** `int`
        -- The number of episodes.
  
    > **epsilon:** `float`
        -- Learning rate.

    > **beta1:** `float`
        -- The exponential decay rate for the first moment estimates.

    > **beta2:** `float`
        -- The exponential decay rate for the second moment estimates.

    > **seed:** `int`
        -- Random seed.

    > **eps:** `float`
        -- Machine epsilon.
    """
    def __init__(
        self,
        savefile=False,
        Adam=False,
        psi0=[],
        ctrl0=[],
        measurement0=[],
        max_episode=300,
        epsilon=0.01,
        beta1=0.90,
        beta2=0.99,
        seed=1234,
        eps=1e-8,
    ):

        Comp.ComprehensiveSystem.__init__(
            self, savefile, psi0, ctrl0, measurement0, seed, eps
        )

        self.Adam = Adam
        self.max_episode = max_episode
        self.epsilon = epsilon
        self.beta1 = beta1
        self.beta2 = beta2
        self.mt = 0.0
        self.vt = 0.0
        self.seed = seed

    def SC(self, W=[], M=[], target="QFIM", LDtype="SLD"):
        """
        Comprehensive optimization of the probe state and control (SC).

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.

        > **M:** `list of matrices`
            -- A set of positive operator-valued measure (POVM). The default measurement 
            is a set of rank-one symmetric informationally complete POVM (SIC-POVM).

        > **target:** `string`
            -- Objective functions for searching the minimum time to reach the given 
            value of the objective function. Options are:  
            "QFIM" (default) -- choose QFI (QFIM) as the objective function.  
            "CFIM" -- choose CFI (CFIM) as the objective function.  
            "HCRB" -- choose HCRB as the objective function.  

        > **LDtype:** `string`
            -- Types of QFI (QFIM) can be set as the objective function. Options are:  
            "SLD" (default) -- QFI (QFIM) based on symmetric logarithmic derivative (SLD).  
            "RLD" -- QFI (QFIM) based on right logarithmic derivative (RLD).  
            "LLD" -- QFI (QFIM) based on left logarithmic derivative (LLD). 
        
        **Note:** AD is only available when target is 'QFIM'.
        """
        if M != []:
            raise ValueError(
                "AD is not available when target is 'CFIM'. Supported methods are 'PSO' and 'DE'.",
            )
        elif target == "HCRB":
            raise ValueError(
                "AD is not available when the target function is HCRB. Supported methods are 'PSO' and 'DE'.",
            )

        if self.Adam:
            self.alg = QJL.QuanEstimation.AD(
                self.max_episode, self.epsilon, self.beta1, self.beta2
            )
        else:
            self.alg = QJL.QuanEstimation.AD(self.max_episode, self.epsilon)

        super().SC(W, M, target, LDtype)



import juliacall
from quanestimation import QJL
import quanestimation.ControlOpt.ControlStruct as Control


class DE_Copt(Control.ControlSystem):
    """
    Attributes
    ----------
    > **savefile:** `bool`
        --Whether or not to save all the control coeffients.  
        If set `True` then the control coefficients and the values of the 
        objective function obtained in all episodes will be saved during 
        the training. If set `False` the control coefficients in the final 
        episode and the values of the objective function in all episodes 
        will be saved.

    > **p_num:** `int`
        -- The number of populations.

    > **ctrl0:** list of arrays
        -- Initial guesses of control coefficients.

    > **max_episode:** `int`
        -- The number of episodes.
  
    > **c:** `float`
        -- Mutation constant.

    > **cr:** `float`
        -- Crossover constant.

    > **seed:** `int`
        -- Random seed.

    > **eps:** `float`
        -- Machine epsilon.

    > **load:** `bool`
        -- Whether or not to load control coefficients in the current location.  
        If set `True` then the program will load control coefficients from 
        "controls.csv" file in the current location and use it as the initial 
        control coefficients.
    """

    def __init__(
        self,
        savefile=False,
        p_num=10,
        ctrl0=[],
        max_episode=1000,
        c=1.0,
        cr=0.5,
        seed=1234,
        eps=1e-8,
        load=False,
    ):

        Control.ControlSystem.__init__(self, savefile, ctrl0, eps, load)

        self.p_num = p_num
        self.max_episode = max_episode
        self.c = c
        self.cr = cr
        self.seed = seed

    def QFIM(self, W=[], LDtype="SLD"):
        r"""
        Choose QFI or $\mathrm{Tr}(WF^{-1})$ as the objective function. 
        In single parameter estimation the objective function is QFI and in 
        multiparameter estimation it will be $\mathrm{Tr}(WF^{-1})$.

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.

        > **LDtype:** `string`
            -- Types of QFI (QFIM) can be set as the objective function. Options are:  
            "SLD" (default) -- QFI (QFIM) based on symmetric logarithmic derivative (SLD).  
            "RLD" -- QFI (QFIM) based on right logarithmic derivative (RLD).  
            "LLD" -- QFI (QFIM) based on left logarithmic derivative (LLD).
        """
        ini_population = (self.ctrl0, )
        self.alg = QJL.DE(
            self.max_episode,
            self.p_num,
            ini_population,
            self.c,
            self.cr,
        )

        super().QFIM(W, LDtype)

    def CFIM(self, M=[], W=[]):
        r"""
        Choose CFI or $\mathrm{Tr}(WI^{-1})$ as the objective function. 
        In single parameter estimation the objective function is CFI and 
        in multiparameter estimation it will be $\mathrm{Tr}(WI^{-1})$.

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.

        > **M:** `list of matrices`
            -- A set of positive operator-valued measure (POVM). The default measurement 
            is a set of rank-one symmetric informationally complete POVM (SIC-POVM).

        **Note:** 
            SIC-POVM is calculated by the Weyl-Heisenberg covariant SIC-POVM fiducial state 
            which can be downloaded from [here](http://www.physics.umb.edu/Research/QBism/
            solutions.html).
        """
        ini_population = (self.ctrl0, )
        self.alg = QJL.DE(
            self.max_episode,
            self.p_num,
            ini_population,
            self.c,
            self.cr,
        )

        super().CFIM(M, W)

    def HCRB(self, W=[]):
        """
        Choose HCRB as the objective function. 

        **Note:** in single parameter estimation, HCRB is equivalent to QFI, please choose 
        QFI as the objective function.

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.
        """
        
        ini_population = (self.ctrl0, )
        self.alg = QJL.DE(
            self.max_episode,
            self.p_num,
            ini_population,
            self.c,
            self.cr,
        )

        super().HCRB(W)

    def mintime(self, f, W=[], M=[], method="binary", target="QFIM", LDtype="SLD"):
        """
        Search of the minimum time to reach a given value of the objective function.

        Parameters
        ----------
        > **f:** `float`
            -- The given value of the objective function.

        > **W:** `matrix`
            -- Weight matrix.

        > **M:** `list of matrices`
            -- A set of positive operator-valued measure (POVM). The default measurement 
            is a set of rank-one symmetric informationally complete POVM (SIC-POVM).

        > **method:** `string`
            -- Methods for searching the minimum time to reach the given value of the 
            objective function. Options are:  
            "binary" (default) -- Binary search (logarithmic search).  
            "forward" -- Forward search from the beginning of time.

        > **target:** `string`
            -- Objective functions for searching the minimum time to reach the given 
            value of the objective function. Options are:<br>
            "QFIM" (default) -- Choose QFI (QFIM) as the objective function.<br>
            "CFIM" -- Choose CFI (CFIM) as the objective function.<br>
            "HCRB" -- Choose HCRB as the objective function.

        > **LDtype:** `string`
            -- Types of QFI (QFIM) can be set as the objective function. Options are:  
            "SLD" (default) -- QFI (QFIM) based on symmetric logarithmic derivative (SLD).  
            "RLD" -- QFI (QFIM) based on right logarithmic derivative (RLD).  
            "LLD" -- QFI (QFIM) based on left logarithmic derivative (LLD).

        **Note:** 
            SIC-POVM is calculated by the Weyl-Heisenberg covariant SIC-POVM fiducial state 
            which can be downloaded from [here](http://www.physics.umb.edu/Research/QBism/
            solutions.html).
        """
        ini_population = (self.ctrl0, )
        self.alg = QJL.DE(
            self.max_episode,
            self.p_num,
            ini_population,
            self.c,
            self.cr,
        )
        
        super().mintime(f, W, M, method, target, LDtype)



from quanestimation import QJL
import quanestimation.StateOpt.StateStruct as State


class RI_Sopt(State.StateSystem):
    """
    Attributes
    ----------
    > **savefile:**  `bool`
        -- Whether or not to save all the states.  
        If set `True` then the states and the values of the 
        objective function obtained in all episodes will be saved during 
        the training. If set `False` the state in the final 
        episode and the values of the objective function in all episodes 
        will be saved.

    > **psi0:** `list of arrays`
        -- Initial guesses of states.

    > **max_episode:** `int`
        -- The number of episodes.

    > **seed:** `int`
        -- Random seed.

    > **eps:** `float`
        -- Machine epsilon.

    > **load:** `bool`
        -- Whether or not to load states in the current location.  
        If set `True` then the program will load state from "states.csv"
        file in the current location and use it as the initial state.
    """

    def __init__(
        self,
        savefile=False,
        psi0=[],
        max_episode=300,
        seed=1234,
        eps=1e-8,
        load=False,
    ):

        State.StateSystem.__init__(self, savefile, psi0, seed, eps, load)

        self.max_episode = max_episode
        self.seed = seed

    def QFIM(self, W=[], LDtype="SLD"):
        r"""
        Choose QFI as the objective function. 

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.

        > **LDtype:** `string`
            -- Types of QFI (QFIM) can be set as the objective function. Only SLD can
            is available here.
        """
        self.alg = QJL.RI(
            self.max_episode,
        )
        if self.dynamics_type != "Kraus":
            raise ValueError("Only the parameterization with Kraus operators is available.")
        
        if LDtype == "SLD":
            super().QFIM(W, LDtype)
        else:
            raise ValueError("Only SLD is available.")

    def CFIM(self, M=[], W=[]):
        """
        Choose CFIM as the objective function. 

        **Note:** CFIM is not available.

        Parameters
        ----------
        > **M:** `list`
            -- POVM.
            
        > **W:** `matrix`
            -- Weight matrix.
        """
        raise ValueError("CFIM is not available.")

    def HCRB(self, W=[]):
        """
        Choose HCRB as the objective function. 

        **Note:** Here HCRB is not available.

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.
        """
        raise ValueError("HCRB is not available.")



from quanestimation.StateOpt.StateStruct import (
    StateSystem,
    StateOpt,
    csv2npy_states,
)
from quanestimation.StateOpt.AD_Sopt import (
    AD_Sopt,
)
from quanestimation.StateOpt.DE_Sopt import (
    DE_Sopt,
)
from quanestimation.StateOpt.PSO_Sopt import (
    PSO_Sopt,
)
# from quanestimation.StateOpt.DDPG_Sopt import (
#     DDPG_Sopt,
# )
from quanestimation.StateOpt.NM_Sopt import (
    NM_Sopt,
)
from quanestimation.StateOpt.RI_Sopt import (
    RI_Sopt,
)

__all__ = [
    "StateSystem",
    "StateOpt",
    "AD_Sopt",
    "DE_Sopt",
    "PSO_Sopt",
    # "DDPG_Sopt",
    "NM_Sopt",
    "RI_Sopt",
    "csv2npy_states",
]



import numpy as np
from numpy.linalg import inv
from scipy.linalg import sqrtm, schur, eigvals
from quanestimation.Common.Common import SIC, suN_generator
from scipy.integrate import quad
from scipy.stats import norm, poisson, rayleigh, gamma

def CFIM(rho, drho, M=[], eps=1e-8):
    r"""
    Calculation of the classical Fisher information matrix for the chosen measurements.

    This function computes the classical Fisher information (CFI) and classical Fisher 
    information matrix (CFIM) for a density matrix. The entry of CFIM $\mathcal{I}$
    is defined as

    $$
    \mathcal{I}_{ab}=\sum_y\frac{1}{p(y|\textbf{x})}[\partial_a p(y|\textbf{x})][\partial_b p(y|\textbf{x})],
    $$

    Symbols: 
        - $p(y|\textbf{x})=\mathrm{Tr}(\rho\Pi_y)$.
        - $\rho$: the parameterized density matrix.

    Args: 
        rho (np.array): 
            Density matrix.
        drho (list): 
            List of derivative matrices of the density matrix on the unknown 
            parameters to be estimated. For example, drho[0] is the derivative 
            matrix on the first parameter.
        M (list, optional): 
            List of positive operator-valued measure (POVM). The default 
            measurement is a set of rank-one symmetric informationally complete POVM (SIC-POVM).
        eps (float, optional): 
            Machine epsilon for numerical stability.

    Returns:
        (float/np.array): 
            For single parameter estimation (the length of drho is equal to one), the output is CFI 
            and for multiparameter estimation (the length of drho is more than one), it returns CFIM.

    Raises:
        TypeError: If drho is not a list.
        TypeError: If M is not a list.   

    Example:
        rho = np.array([[0.5, 0], [0, 0.5]]);

        drho = [np.array([[1, 0], [0, -1]])];

        cfim = CFIM(rho, drho);     
    
    Notes: 
        SIC-POVM is calculated by the Weyl-Heisenberg covariant SIC-POVM fiducial state 
        which can be downloaded from [here](https://www.physics.umb.edu/Research/QBism/solutions.html).
    """

    if not isinstance(drho, list):
        raise TypeError("Please make sure drho is a list!")

    if not M:
        M = SIC(len(rho[0]))
    else:
        if not isinstance(M, list):
            raise TypeError("Please make sure M is a list!")

    num_measurements = len(M)
    num_params = len(drho)
    cfim_res = np.zeros([num_params, num_params])
    
    for i in range(num_measurements):
        povm_element = M[i]
        p = np.real(np.trace(rho @ povm_element))
        c_add = np.zeros([num_params, num_params])
        
        if p > eps:
            for param_i in range(num_params):
                drho_i = drho[param_i]
                dp_i = np.real(np.trace(drho_i @ povm_element))
                
                for param_j in range(param_i, num_params):
                    drho_j = drho[param_j]
                    dp_j = np.real(np.trace(drho_j @ povm_element))
                    c_add[param_i][param_j] = np.real(dp_i * dp_j / p)
                    c_add[param_j][param_i] = np.real(dp_i * dp_j / p)
                    
        cfim_res += c_add

    if num_params == 1:
        return cfim_res[0][0]
    else:
        return cfim_res


def FIM(p, dp, eps=1e-8):
    r"""
    Calculation of the classical Fisher information matrix (CFIM) for a given probability distributions.

    This function computes the classical Fisher information matrix (CFIM) for a given probability 
    distributions. The entry of FIM $I$ is defined as

    $$
    I_{ab}=\sum_{y}\frac{1}{p_y}[\partial_a p_y][\partial_b p_y],
    $$

    Symbols: 
        - $\{p_y\}$: a set of the discrete probability distribution.

    Args: 
        p (np.array): 
            The probability distribution.
        dp (list): 
            Derivatives of the probability distribution on the unknown parameters to 
            be estimated. For example, dp[0] is the derivative vector on the first parameter.
        eps (float, optional): 
            Machine epsilon.

    Returns:
        (float/np.array): 
            For single parameter estimation (the length of drho is equal to one), the output is CFI 
            and for multiparameter estimation (the length of drho is more than one), it returns CFIM.
    """

    num_params = len(dp)
    num_measurements = len(p)
    fim_matrix = np.zeros([num_params, num_params])
    
    for outcome_idx in range(num_measurements):
        p_value = p[outcome_idx]
        fim_add = np.zeros([num_params, num_params])
        
        if p_value > eps:
            for param_i in range(num_params):
                dp_i = dp[param_i][outcome_idx]
                
                for param_j in range(param_i, num_params):
                    dp_j = dp[param_j][outcome_idx]
                    term = np.real(dp_i * dp_j / p_value)
                    fim_add[param_i][param_j] = term
                    fim_add[param_j][param_i] = term
                    
        fim_matrix += fim_add

    if num_params == 1:
        return fim_matrix[0][0]
    else:
        return fim_matrix

def FI_Expt(data_true, data_shifted, delta_x, ftype="norm"):
    """
    Calculate the classical Fisher information (CFI) based on experimental data.

    Args:
        data_true (np.array): 
            Experimental data obtained at the true parameter value.
        data_shifted (np.array): 
            Experimental data obtained at parameter value shifted by delta_x.
        delta_x (float): 
            Small known parameter shift.
        ftype (str, optional): 
            Probability distribution of the data. Options:  
                - "norm": normal distribution (default).  
                - "gamma": gamma distribution.  
                - "rayleigh": Rayleigh distribution.  
                - "poisson": Poisson distribution.  

    Returns: 
        (float): 
            Classical Fisher information

    Raises:
        ValueError: 
            If `ftype` is not one of the supported types ("norm", "poisson", "gamma", "rayleigh").    

    Notes:
        The current implementation may be unstable and is subject to future modification.
    """
    fidelity = 0.0
    if ftype == "norm":
        mu_true, std_true = norm.fit(data_true)
        mu_shifted, std_shifted = norm.fit(data_shifted)
        f_function = lambda x: np.sqrt(
            norm.pdf(x, mu_true, std_true) * norm.pdf(x, mu_shifted, std_shifted)
        )
        fidelity, _ = quad(f_function, -np.inf, np.inf)
        
    elif ftype == "gamma":
        a_true, alpha_true, beta_true = gamma.fit(data_true)
        a_shifted, alpha_shifted, beta_shifted = gamma.fit(data_shifted)
        f_function = lambda x: np.sqrt(
            gamma.pdf(x, a_true, alpha_true, beta_true) *
            gamma.pdf(x, a_shifted, alpha_shifted, beta_shifted)
        )
        fidelity, _ = quad(f_function, 0., np.inf)
        
    elif ftype == "rayleigh":
        mean_true, var_true = rayleigh.fit(data_true)
        mean_shifted, var_shifted = rayleigh.fit(data_shifted)
        f_function = lambda x: np.sqrt(
            rayleigh.pdf(x, mean_true, var_true) *
            rayleigh.pdf(x, mean_shifted, var_shifted)
        )
        fidelity, _ = quad(f_function, -np.inf, np.inf)
        
    elif ftype == "poisson":
        k_max = max(max(data_true) + 1, max(data_shifted) + 1)
        k_values = np.arange(k_max)
        p_true = poisson.pmf(k_values, np.mean(data_true))
        p_shifted = poisson.pmf(k_values, np.mean(data_shifted))
        p_true /= np.sum(p_true)
        p_shifted /= np.sum(p_shifted)
        fidelity = np.sum(np.sqrt(p_true * p_shifted))
        
    else:
        valid_types = ["norm", "poisson", "gamma", "rayleigh"]
        raise ValueError(
            f"Invalid distribution type: '{ftype}'. "
            f"Supported types are: {', '.join(valid_types)}"
        )
    
    fisher_information = 8 * (1 - fidelity) / delta_x**2
    return fisher_information


def SLD(rho, drho, rep="original", eps=1e-8):
    r"""
    Calculation of the symmetric logarithmic derivative (SLD) for a density matrix.

    This function computes the SLD operator $L_a$, which is determined by

    $$
    \partial_{a}\rho=\frac{1}{2}(\rho L_{a}+L_{a}\rho)
    $$

    with $\rho$ the parameterized density matrix. The entries of SLD can be calculated as 

    $$
    \langle\lambda_i|L_{a}|\lambda_j\rangle=\frac{2\langle\lambda_i| \partial_{a}\rho |\lambda_j\rangle}{\lambda_i+\lambda_j
    $$

    for $\lambda_i~(\lambda_j) \neq 0$. If $\lambda_i=\lambda_j=0$, the entry of SLD is set to be zero.

    Args:
        rho (np.array): 
            Density matrix.
        drho (list): 
            Derivatives of the density matrix on the unknown parameters to be 
            estimated. For example, drho[0] is the derivative vector on the first parameter.
        rep (str, optional): 
            The basis for the SLDs. Options:  
                - "original" (default): basis same as input density matrix  
                - "eigen": basis same as eigenspace of density matrix
        eps (float, optional): 
            Machine epsilon.

    Returns:
        (np.array/list): 
            For single parameter estimation (i.e., length of `drho` equals 1), returns a matrix.  
            For multiparameter estimation (i.e., length of `drho` is larger than 1), returns a list of matrices.

    Raises:
        TypeError: If `drho` is not a list.  
        ValueError: If `rep` has invalid value. 
    """

    if not isinstance(drho, list):
        raise TypeError("drho must be a list of derivative matrices")

    num_params = len(drho)
    dim = len(rho)
    slds = [None] * num_params

    purity = np.trace(rho @ rho)

    # Handle pure state case
    if np.abs(1 - purity) < eps:
        sld_original = [2 * d for d in drho]
        
        for i in range(num_params):
            if rep == "original":
                slds[i] = sld_original[i]
            elif rep == "eigen":
                eigenvalues, eigenvectors = np.linalg.eig(rho)
                eigenvalues = np.real(eigenvalues)
                slds[i] = eigenvectors.conj().T @ sld_original[i] @ eigenvectors
            else:
                valid_reps = ["original", "eigen"]
                raise ValueError(f"Invalid rep value: '{rep}'. Valid options: {valid_reps}")
        
        return slds[0] if num_params == 1 else slds

    # Handle mixed state case
    eigenvalues, eigenvectors = np.linalg.eig(rho)
    eigenvalues = np.real(eigenvalues)
    
    for param_idx in range(num_params):
        sld_eigenbasis = np.zeros((dim, dim), dtype=np.complex128)
        
        for i in range(dim):
            for j in range(dim):
                if eigenvalues[i] + eigenvalues[j] > eps:
                    # Calculate matrix element in eigenbasis
                    numerator = 2 * (eigenvectors[:, i].conj().T @ drho[param_idx] @ eigenvectors[:, j])
                    sld_eigenbasis[i, j] = numerator / (eigenvalues[i] + eigenvalues[j])
        
        # Handle any potential infinities
        sld_eigenbasis[np.isinf(sld_eigenbasis)] = 0.0
        
        # Transform to requested basis
        if rep == "original":
            slds[param_idx] = eigenvectors @ sld_eigenbasis @ eigenvectors.conj().T
        elif rep == "eigen":
            slds[param_idx] = sld_eigenbasis
        else:
            valid_reps = ["original", "eigen"]
            raise ValueError(f"Invalid rep value: '{rep}'. Valid options: {valid_reps}")
    
    return slds[0] if num_params == 1 else slds


def RLD(rho, drho, rep="original", eps=1e-8):
    r"""
    Calculation of the right logarithmic derivative (RLD) for a density matrix.
    The RLD operator $\mathcal{R}_a$ is defined by

    $$
    \partial_{a}\rho=\rho \mathcal{R}_a
    $$

    with $\rho$ the parameterized density matrix. The entries of RLD can be calculated as 

    $$
    \langle\lambda_i| \mathcal{R}_{a} |\lambda_j\rangle=\frac{1}{\lambda_i}\langle\lambda_i| 
    \partial_a\rho |\lambda_j\rangle 
    $$

    for $\lambda_i\neq 0$.

    Args:
        rho (np.array): 
            Density matrix.  
        drho (list):  
            Derivatives of the density matrix on the unknown parameters to be 
            estimated. For example, drho[0] is the derivative vector on the first parameter.
        rep (str, optional): 
            The basis for the RLD(s). Options:  
                - "original" (default): basis same as input density matrix.  
                - "eigen": basis same as eigenspace of density matrix.
        eps (float, optional): 
            Machine epsilon.

    Returns:
        (np.array/list): 
            For single parameter estimation (i.e., length of `drho` equals 1), returns a matrix.  
            For multiparameter estimation (i.e., length of `drho` is larger than 1), returns a list of matrices.

    Raises:
        TypeError: If `drho` is not a list.
        ValueError: If `rep` has invalid value or RLD doesn't exist.
    """
    
    if not isinstance(drho, list):
        raise TypeError("drho must be a list of derivative matrices")

    num_params = len(drho)
    dim = len(rho)
    rld_list = [None] * num_params

    eigenvalues, eigenvectors = np.linalg.eig(rho)
    eigenvalues = np.real(eigenvalues)
    
    for param_idx in range(num_params):
        rld_eigenbasis = np.zeros((dim, dim), dtype=np.complex128)
        
        for i in range(dim):
            for j in range(dim):
                # Calculate matrix element in eigenbasis
                element = (
                    eigenvectors[:, i].conj().T 
                    @ drho[param_idx] 
                    @ eigenvectors[:, j]
                )
                
                if np.abs(eigenvalues[i]) > eps:
                    rld_eigenbasis[i, j] = element / eigenvalues[i]
                else:
                    if np.abs(element) > eps:
                        raise ValueError(
                            "RLD does not exist. It only exists when the support of "
                            "drho is contained in the support of rho."
                        )
        
        # Handle any potential infinities
        rld_eigenbasis[np.isinf(rld_eigenbasis)] = 0.0
        
        # Transform to requested basis
        if rep == "original":
            rld_list[param_idx] = (
                eigenvectors 
                @ rld_eigenbasis 
                @ eigenvectors.conj().T
            )
        elif rep == "eigen":
            rld_list[param_idx] = rld_eigenbasis
        else:
            valid_reps = ["original", "eigen"]
            raise ValueError(
                f"Invalid rep value: '{rep}'. Valid options: {', '.join(valid_reps)}"
            )
    
    return rld_list[0] if num_params == 1 else rld_list


def LLD(rho, drho, rep="original", eps=1e-8):
    r"""
    Calculation of the left logarithmic derivative (LLD) for a density matrix $\rho$.

    The LLD operator $\mathcal{R}_a^{\dagger}$ is defined by

    $$
    \partial_{a}\rho=\mathcal{R}_a^{\dagger}\rho.
    $$

    The entries of LLD can be calculated as 

    $$
    \langle\lambda_i| \mathcal{R}_{a}^{\dagger} |\lambda_j\rangle=\frac{1}{\lambda_j}\langle\lambda_i| 
    \partial_a\rho |\lambda_j\rangle 
    $$

    for $\lambda_j\neq 0$.

    Args: 
        rho (np.array): 
            Density matrix.
        drho (list): 
            Derivatives of the density matrix on the unknown parameters to be estimated. 
            For example, drho[0] is the derivative vector on the first parameter.
        rep (str, optional): 
            The basis for the LLD(s). Options:  
                - "original" (default): basis same as input density matrix.  
                - "eigen": basis same as eigenspace of density matrix.
        eps (float, optional): 
            Machine epsilon.

    Returns:
        (np.array/list): 
            For single parameter estimation (i.e., length of `drho` equals 1), returns a matrix.  
            For multiparameter estimation (i.e., length of `drho` is larger than 1), returns a list of matrices.

    Raises:
        TypeError: If `drho` is not a list.  
        ValueError: If `rep` has invalid value or LLD doesn't exist.  
    """

    if not isinstance(drho, list):
        raise TypeError("drho must be a list of derivative matrices")

    param_num = len(drho)
    dim = len(rho)
    lld_list = [None] * param_num

    eigenvalues, eigenvectors = np.linalg.eig(rho)
    eigenvalues = np.real(eigenvalues)
    
    for param_idx in range(param_num):
        lld_eigenbasis = np.zeros((dim, dim), dtype=np.complex128)
        
        for i in range(dim):
            for j in range(dim):
                # Calculate matrix element in eigenbasis
                element = (
                    eigenvectors[:, i].conj().T 
                    @ drho[param_idx] 
                    @ eigenvectors[:, j]
                )
                
                if np.abs(eigenvalues[j]) > eps:
                    lld_eigenbasis[i, j] = element / eigenvalues[j]
                else:
                    if np.abs(element) > eps:
                        raise ValueError(
                            "LLD does not exist. It only exists when the support of "
                            "drho is contained in the support of rho."
                        )
        
        # Handle any potential infinities
        lld_eigenbasis[np.isinf(lld_eigenbasis)] = 0.0
        
        # Transform to requested basis
        if rep == "original":
            lld_list[param_idx] = (
                eigenvectors 
                @ lld_eigenbasis 
                @ eigenvectors.conj().T
            )
        elif rep == "eigen":
            lld_list[param_idx] = lld_eigenbasis
        else:
            valid_reps = ["original", "eigen"]
            raise ValueError(
                f"Invalid rep value: '{rep}'. Valid options: {', '.join(valid_reps)}"
            )
    
    return lld_list[0] if param_num == 1 else lld_list


def QFIM(rho, drho, LDtype="SLD", exportLD=False, eps=1e-8):
    r"""
    Calculate the quantum Fisher information (QFI) and quantum Fisher 
    information matrix (QFIM) for all types.

    The entry of QFIM $\mathcal{F}$ is defined as:

    $$
    \mathcal{F}_{ab}=\frac{1}{2}\mathrm{Tr}(\rho\{L_a, L_b\})
    $$

    with $L_a, L_b$ being SLD operators.

    Alternatively:

    $$
    \mathcal{F}_{ab}=\mathrm{Tr}(\rho \mathcal{R}_a \mathcal{R}^{\dagger}_b)
    $$

    with $\mathcal{R}_a$ being the RLD or LLD operator.

    Args:
        rho (np.array): 
            Density matrix.
        drho (list): 
            Derivatives of the density matrix with respect to the unknown parameters. 
            Each element in the list is a matrix of the same dimension as `rho` and 
            represents the partial derivative of the density matrix with respect to 
            one parameter. For example, `drho[0]` is the derivative with respect to 
            the first parameter.
        LDtype (str, optional): 
            Specifies the type of logarithmic derivative to use for QFI/QFIM calculation:  
                - "SLD": Symmetric Logarithmic Derivative (default).  
                - "RLD": Right Logarithmic Derivative.  
                - "LLD": Left Logarithmic Derivative.  
        exportLD (bool, optional): 
            Whether to export the values of logarithmic derivatives.  
        eps (float, optional): 
            Machine epsilon.  

    Returns:
        (float/np.array): 
            For single parameter estimation (i.e., length of `drho` equals 1), returns QFI.  
            For multiparameter estimation (i.e., length of `drho` is larger than 1), returns QFIM.  

    Raises:
        TypeError: If `drho` is not a list.
        ValueError: If `LDtype` is not one of the supported types ("SLD", "RLD", "LLD").        
    """

    if not isinstance(drho, list):
        raise TypeError("drho must be a list of derivative matrices")

    num_params = len(drho)
    qfim_result = None
    log_derivatives = None

    # Single parameter estimation
    if num_params == 1:
        if LDtype == "SLD":
            sld = SLD(rho, drho, eps=eps)
            anticommutator = sld @ sld + sld @ sld
            qfim_result = np.real(0.5 * np.trace(rho @ anticommutator))
        elif LDtype == "RLD":
            rld = RLD(rho, drho, eps=eps)
            qfim_result = np.real(np.trace(rho @ rld @ rld.conj().T))
        elif LDtype == "LLD":
            lld = LLD(rho, drho, eps=eps)
            qfim_result = np.real(np.trace(rho @ lld @ lld.conj().T))
        else:
            valid_types = ["SLD", "RLD", "LLD"]
            raise ValueError(
                f"Invalid LDtype: '{LDtype}'. Supported types: {', '.join(valid_types)}"
            )
        log_derivatives = sld if LDtype == "SLD" else rld if LDtype == "RLD" else lld

    # Multiparameter estimation
    else:
        if LDtype == "SLD":
            qfim_result = np.zeros((num_params, num_params))
            sld_list = SLD(rho, drho, eps=eps)
            for i in range(num_params):
                for j in range(i, num_params):
                    anticommutator = sld_list[i] @ sld_list[j] + sld_list[j] @ sld_list[i]
                    qfim_result[i, j] = np.real(0.5 * np.trace(rho @ anticommutator))
                    qfim_result[j, i] = qfim_result[i, j]
            log_derivatives = sld_list

        elif LDtype == "RLD":
            qfim_result = np.zeros((num_params, num_params), dtype=np.complex128)
            rld_list = RLD(rho, drho, eps=eps)
            for i in range(num_params):
                for j in range(i, num_params):
                    term = np.trace(rho @ rld_list[i] @ rld_list[j].conj().T)
                    qfim_result[i, j] = term
                    qfim_result[j, i] = term.conj()
            log_derivatives = rld_list

        elif LDtype == "LLD":
            qfim_result = np.zeros((num_params, num_params), dtype=np.complex128)
            lld_list = LLD(rho, drho, eps=eps)
            for i in range(num_params):
                for j in range(i, num_params):
                    term = np.trace(rho @ lld_list[i] @ lld_list[j].conj().T)
                    qfim_result[i, j] = term
                    qfim_result[j, i] = term.conj()
            log_derivatives = lld_list

        else:
            valid_types = ["SLD", "RLD", "LLD"]
            raise ValueError(
                f"Invalid LDtype: '{LDtype}'. Supported types: {', '.join(valid_types)}"
            )

    if exportLD:
        return qfim_result, log_derivatives
    return qfim_result


def QFIM_Kraus(rho0, K, dK, LDtype="SLD", exportLD=False, eps=1e-8):
    r"""
    Calculation of the quantum Fisher information (QFI) and quantum Fisher 
    information matrix (QFIM) for a quantum channel described by Kraus operators.

    The quantum channel is given by
    
    $$
    \rho=\sum_{i} K_i \rho_0 K_i^{\dagger},
    $$

    where $\rho_0$ is the initial state and $\{K_i\}$ are the Kraus operators.

    The derivatives of the density matrix $\partial_a\rho$ are calculated from the 
    derivatives of the Kraus operators $\{\partial_a K_i\}$ as
    
    $$
    \partial_a\rho=\sum_{i}\left[(\partial_a K_i)\rho_0 K_i^{\dagger}+K_i\rho_0(\partial_a K_i)^{\dagger}\right].
    $$

    Then the QFI (QFIM) is calculated via the function `QFIM` with the evolved state 
    $\rho$ and its derivatives $\{\partial_a\rho\}$.

    Args:
        rho0 (np.array): 
            Initial density matrix.
        K (list): 
            Kraus operators.
        dK (list): 
            Derivatives of the Kraus operators. It is a nested list where the first index 
            corresponds to the parameter and the second index corresponds to the Kraus operator index. 
            For example, `dK[0][1]` is the derivative of the second Kraus operator with respect 
            to the first parameter.
        LDtype (str, optional): 
            Types of QFI (QFIM) can be set as the objective function. Options:  
                - "SLD" (default): QFI (QFIM) based on symmetric logarithmic derivative.  
                - "RLD": QFI (QFIM) based on right logarithmic derivative.  
                - "LLD": QFI (QFIM) based on left logarithmic derivative.  
        exportLD (bool, optional): 
            Whether to export the values of logarithmic derivatives.  
        eps (float, optional): 
            Machine epsilon.  

    Returns:
        (float/np.array): 
            For single parameter estimation (the length of dK is equal to one), the output is QFI 
            and for multiparameter estimation (the length of dK is more than one), it returns QFIM.
    """

    # Transpose dK: from [parameters][operators] to [operators][parameters]
    dK_transposed = [
        [dK[i][j] for i in range(len(K))] 
        for j in range(len(dK[0]))
    ]
    
    # Compute the evolved density matrix
    rho = sum(Ki @ rho0 @ Ki.conj().T for Ki in K)
    
    # Compute the derivatives of the density matrix
    drho = [
        sum(
            dKi @ rho0 @ Ki.conj().T + Ki @ rho0 @ dKi.conj().T
            for Ki, dKi in zip(K, dKj)
        )
        for dKj in dK_transposed
    ]
    
    return QFIM(rho, drho, LDtype=LDtype, exportLD=exportLD, eps=eps)


def QFIM_Bloch(r, dr, eps=1e-8):
    r"""
    Calculation of the quantum Fisher information (QFI) and quantum Fisher 
    information matrix (QFIM) in Bloch representation.

    The Bloch vector representation of a quantum state is defined as
    
    $$
    \rho = \frac{1}{d}\left(\mathbb{I} + \sum_{i=1}^{d^2-1} r_i \lambda_i\right),
    $$
    
    where $\lambda_i$ are the generators of SU(d) group.

    Args:
        r (np.array): 
            Parameterized Bloch vector.
        dr (list): 
            Derivatives of the Bloch vector with respect to the unknown parameters. 
            Each element in the list is a vector of the same length as `r` and 
            represents the partial derivative of the Bloch vector with respect to 
            one parameter. For example, `dr[0]` is the derivative with respect to 
            the first parameter.
        eps (float, optional): 
            Machine epsilon.  

    Returns:
        (float/np.array): 
            For single parameter estimation (the length of `dr` is equal to one), 
            the output is QFI and for multiparameter estimation (the length of `dr` 
            is more than one), it returns QFIM.

    Raises:
        TypeError: If `dr` is not a list.  
        ValueError: If the dimension of the Bloch vector is invalid.  
    """

    if not isinstance(dr, list):
        raise TypeError("dr must be a list of derivative vectors")

    num_params = len(dr)
    qfim_result = np.zeros((num_params, num_params))

    # Calculate dimension from Bloch vector length
    dim_float = np.sqrt(len(r) + 1)
    if dim_float.is_integer():
        dim = int(dim_float)
    else:
        raise ValueError("Invalid Bloch vector dimension")

    # Get SU(N) generators
    lambda_generators = suN_generator(dim)

    # Handle single-qubit system
    if dim == 2:
        r_norm = np.linalg.norm(r) ** 2
        
        # Pure state case
        if np.abs(r_norm - 1.0) < eps:
            for i in range(num_params):
                for j in range(i, num_params):
                    qfim_result[i, j] = np.real(np.inner(dr[i], dr[j]))
                    qfim_result[j, i] = qfim_result[i, j]
        # Mixed state case
        else:
            for i in range(num_params):
                for j in range(i, num_params):
                    term1 = np.inner(dr[i], dr[j])
                    term2 = (np.inner(r, dr[i]) * np.inner(r, dr[j])) / (1 - r_norm)
                    qfim_result[i, j] = np.real(term1 + term2)
                    qfim_result[j, i] = qfim_result[i, j]
    # Handle higher-dimensional systems
    else:
        # Reconstruct density matrix from Bloch vector
        rho = np.eye(dim, dtype=np.complex128) / dim
        for idx in range(dim**2 - 1):
            coeff = np.sqrt(dim * (dim - 1) / 2) * r[idx] / dim
            rho += coeff * lambda_generators[idx]

        # Calculate G matrix
        G = np.zeros((dim**2 - 1, dim**2 - 1), dtype=np.complex128)
        for i in range(dim**2 - 1):
            for j in range(i, dim**2 - 1):
                anticommutator = (
                    lambda_generators[i] @ lambda_generators[j] + 
                    lambda_generators[j] @ lambda_generators[i]
                )
                G[i, j] = 0.5 * np.trace(rho @ anticommutator)
                G[j, i] = G[i, j]

        # Calculate matrix for inversion
        r_vec = np.array(r).reshape(len(r), 1)
        mat = G * dim / (2 * (dim - 1)) - r_vec @ r_vec.T
        mat_inv = inv(mat)

        # Calculate QFIM
        for i in range(num_params):
            for j in range(i, num_params):
                dr_i = np.array(dr[i]).reshape(1, len(r))
                dr_j = np.array(dr[j]).reshape(len(r), 1)
                qfim_result[i, j] = np.real(dr_i @ mat_inv @ dr_j)[0, 0]
                qfim_result[j, i] = qfim_result[i, j]

    return qfim_result[0, 0] if num_params == 1 else qfim_result


def QFIM_Gauss(R, dR, D, dD):
    r"""
    Calculation of the quantum Fisher information (QFI) and quantum 
    Fisher information matrix (QFIM) for Gaussian states.

    The Gaussian state is characterized by its first-order moment (displacement vector) 
    and second-order moment (covariance matrix). The QFIM is calculated using the 
    method described in [1].

    Args:
        R (np.array): 
            First-order moment (displacement vector).
        dR (list): 
            Derivatives of the first-order moment with respect to the unknown parameters. 
            Each element in the list is a vector of the same length as `R` and represents the partial 
            derivative of the displacement vector with respect to one parameter. For example, `dR[0]` 
            is the derivative with respect to the first parameter.
        D (np.array): 
            Second-order moment (covariance matrix).
        dD (list): 
            Derivatives of the second-order moment with respect to the unknown parameters. 
            Each element in the list is a matrix of the same dimension as `D` and 
            represents the partial derivative of the covariance matrix with respect to 
            one parameter. For example, `dD[0]` is the derivative with respect to 
            the first parameter.

    Returns:
        (float/np.array): 
            For single parameter estimation (the length of `dR` is equal to one), 
            the output is QFI and for multiparameter estimation (the length of `dR` 
            is more than one), it returns QFIM.

    Notes:
        This function follows the approach from:
        [1] Monras, A., Phase space formalism for quantum estimation of Gaussian states, arXiv:1303.3682 (2013).
    """

    num_params = len(dR)
    m = len(R) // 2  # Number of modes
    qfim = np.zeros((num_params, num_params))

    # Compute the covariance matrix from the second-order moments and displacement
    cov_matrix = np.array(
        [
            [D[i][j] - R[i] * R[j] for j in range(2 * m)]
            for i in range(2 * m)
        ]
    )

    # Compute the derivatives of the covariance matrix
    dcov = []
    for k in range(num_params):
        dcov_k = np.zeros((2 * m, 2 * m))
        for i in range(2 * m):
            for j in range(2 * m):
                dcov_k[i, j] = dD[k][i][j] - dR[k][i] * R[j] - R[i] * dR[k][j]
        dcov.append(dcov_k)

    # Compute the square root of the covariance matrix
    cov_sqrt = sqrtm(cov_matrix)

    # Define the symplectic matrix J for m modes
    J_block = np.array([[0, 1], [-1, 0]])
    J = np.kron(J_block, np.eye(m))

    # Compute the matrix B = cov_sqrt @ J @ cov_sqrt
    B = cov_sqrt @ J @ cov_sqrt

    # Permutation matrix to rearrange the basis
    P = np.eye(2 * m)
    # Rearrange the basis: first all q's then all p's
    P = np.vstack([P[::2], P[1::2]])

    # Schur decomposition of B
    _, Q = schur(B)
    eigenvalues = eigvals(B)
    # Extract the imaginary parts of every other eigenvalue
    c = eigenvalues[::2].imag

    # Diagonal matrix with entries 1/sqrt(c_i) for each mode
    diag_inv_sqrt = np.diagflat(1.0 / np.sqrt(c))

    # Construct the matrix S
    temp1 = J @ cov_sqrt @ Q
    temp2 = P @ np.kron(np.array([[0, 1], [-1, 0]]), -diag_inv_sqrt)
    S = inv(temp1 @ temp2).T @ P.T

    # Define the basis matrices for the Gaussian representation
    sigma_x = np.array([[0.0, 1.0], [1.0, 0.0]])
    sigma_y = np.array([[0.0, -1.0j], [1.0j, 0.0]])
    sigma_z = np.array([[1.0, 0.0], [0.0, -1.0]])
    identity = np.eye(2)
    a_gauss = [1j * sigma_y, sigma_z, identity, sigma_x]

    # Construct the basis matrices for the m-mode system
    es = []
    for i in range(m):
        row = []
        for j in range(m):
            e_ij = np.eye(1, m * m, m * i + j).reshape(m, m)
            row.append(e_ij)
        es.append(row)

    # As: a list of two-mode basis matrices for each of the four types
    As = []
    for a in a_gauss:
        A_type = []
        for i in range(m):
            for j in range(m):
                A_ij = np.kron(es[i][j], a) / np.sqrt(2)
                A_type.append(A_ij)
        As.append(A_type)

    # Compute the coefficients g for each parameter and each basis matrix
    g = []
    for k in range(num_params):
        g_k = []
        for A_type in As:
            g_type = []
            for A_mat in A_type:
                term = np.trace(inv(S) @ dcov[k] @ inv(S.T) @ A_mat.T)
                g_type.append(term)
            g_k.append(g_type)
        g.append(g_k)

    # Initialize the matrices G for each parameter
    G_matrices = [np.zeros((2 * m, 2 * m), dtype=np.complex128) for _ in range(num_params)]

    # Construct the matrices G for each parameter
    for k in range(num_params):
        for i in range(m):
            for j in range(m):
                for l in range(4):  # For each of the four types
                    denom = 4 * c[i] * c[j] + (-1) ** (l + 1)
                    A_l_ij = As[l][i * m + j]
                    G_matrices[k] += np.real(
                        g[k][l][i * m + j] / denom * inv(S.T) @ A_l_ij @ inv(S)
                    )

    # Compute the QFIM
    for i in range(num_params):
        for j in range(num_params):
            term1 = np.trace(G_matrices[i] @ dcov[j])
            term2 = dR[i] @ inv(cov_matrix) @ dR[j]
            qfim[i, j] = np.real(term1 + term2)

    if num_params == 1:
        return qfim[0, 0]
    else:
        return qfim



from quanestimation.Parameterization.GeneralDynamics import (
    Lindblad,
)
from quanestimation.Parameterization.NonDynamics import (
    Kraus,
)

__all__ = [
    "Lindblad",
    "secondorder_derivative",
    "Kraus", 
]



import pytest
import numpy as np
from quanestimation.AsymptoticBound.AnalogCramerRao import HCRB, NHB

def test_HCRB_NHB():
    """
    Test Holevo Cramer-Rao bound (HCRB) and Nagaoka-Hayashi bound (NHB).
    Checks HCRB/NHB calculation for parameterized quantum state.
    """
    # Parameterized state
    theta = np.pi / 3
    phi = 0.0
    psi = np.array(
        [[np.cos(theta / 2)], 
         [np.sin(theta / 2) * np.exp(1j * phi)]]
    )
    rho = psi @ psi.conj().T
    W = np.array(
        [[1.0, 0], 
         [0, np.sin(theta) ** 2]]
    )
    
    # Density matrix derivatives
    drho_theta = np.array([
        [-np.sin(theta) / 2, np.cos(theta) * np.exp(-1j * phi) / 2],
        [np.cos(theta) * np.exp(1j * phi) / 2, np.sin(theta) / 2]
    ])
    drho_phi = np.array([
        [0, -1j * np.sin(theta) / 2 * np.exp(-1j * phi)],
        [1j * np.sin(theta) / 2 * np.exp(1j * phi), 0]
    ])
    drho = [drho_theta, drho_phi]
    
    # Calculate bounds
    result_hcrb = HCRB(rho, drho, W)
    result_nhb = NHB(rho, drho, W)
    
    # Expected results
    expected_hcrb = 4
    expected_nhb = 4
    
    # Verify calculations
    assert np.allclose(result_hcrb, expected_hcrb)
    assert np.allclose(result_nhb, expected_nhb)


def test_HCRB_NHB_invalid_input():
    """
    Test HCRB/NHB with invalid input.
    Checks TypeError raised for invalid density matrix.
    """
    # Invalid density matrix
    rho = np.array(
        [[0.5, 0.5], 
         [0.5, 0.5]]
    )
    drho_invalid = np.array(
        [[1, 2], 
         [2, 1]]
    )  # Dummy derivative
    W = np.array(
        [[1.0, 0], 
         [0, 1]]
    )  # Dummy weight matrix
    
    with pytest.raises(TypeError):
        HCRB(rho, drho_invalid, W)
    with pytest.raises(TypeError):
        NHB(rho, drho_invalid, W)

def test_HCRB_print(capfd):
    """
    Test print statements in HCRB.
    Checks correct messages printed for special cases.
    """
    rho = np.array(
        [[0.5, 0.5], 
         [0.5, 0.5]]
    )
    
    # Single parameter case
    drho1 = [np.array(
        [[1, 2], 
         [2, 1]]
    )]
    W = np.array(
        [[1.0, 0], 
         [0, 1]]
    )
    HCRB(rho, drho1, W)
    out, _ = capfd.readouterr()
    assert  "In single parameter scenario, HCRB is equivalent to QFI. Returning QFI value." in out
    
    # Rank-one weight matrix case
    drho2 = [
        np.array([[1, 0], [0, 0]]), 
        np.array([[0, 1], [1, 0]])
    ]
    W1 = np.array(
        [[1.0, 0], 
         [0, 0]]
    )  # Rank-one weight matrix
    HCRB(rho, drho2, W1)
    out, _ = capfd.readouterr()
    assert "For rank-one weight matrix, HCRB is equivalent to QFIM. Returning Tr(W @ inv(QFIM))." in out



import numpy as np
import os
import copy
from scipy.sparse import csc_matrix, csr_matrix
from sympy import Matrix, GramSchmidt
from itertools import product
import juliacall


def load_julia():
    """
    Load Julia and initialize QuanEstimation module.
    
    Returns:
        jl.Main.QuanEstimation: Julia module for quantum estimation
    """
    jl = juliacall.newmodule("QuanEstimation")
    jl.Main.seval("using QuanEstimation, PythonCall")
    return jl.Main.QuanEstimation


def mat_vec_convert(A):
    """
    Convert between matrix and vector representations.
    
    Args:
        A (np.array): 
            Input matrix or vector. 
        
    Returns:
        (np.array): 
            Converted matrix or vector.
    """
    if A.shape[1] == 1:
        dim = int(np.sqrt(len(A)))
        return A.reshape([dim, dim])
    else:
        return A.reshape([len(A) ** 2, 1])


def suN_unsorted(n):
    """
    Generate unsorted SU(N) generators.
    
    Args:
        n (float): 
            Dimension of the system.
        
    Returns:
        (tuple): 
            Tuple of symmetric, antisymmetric, and diagonal generators.
    """
    U, V, W = [], [], []
    for i in range(1, n):
        for j in range(0, i):
            U_tp = csc_matrix(
                ([1.0, 1.0], ([i, j], [j, i])), 
                shape=(n, n)
            ).toarray()
            V_tp = csc_matrix(
                ([1.0j, -1.0j], ([i, j], [j, i])), 
                shape=(n, n)
            ).toarray()
            U.append(U_tp)
            V.append(V_tp)

    diag = []
    for i in range(n - 1):
        mat_tp = csc_matrix(
            ([1, -1], ([i, i + 1], [i, i + 1])), 
            shape=(n, n)
        ).toarray()
        diag_tp = np.diagonal(mat_tp)
        diag.append(Matrix(diag_tp))
        
    W_gs = GramSchmidt(diag, True)
    for k in range(len(W_gs)):
        W_tp = np.fromiter(W_gs[k], dtype=complex)
        W.append(np.sqrt(2) * np.diag(W_tp))

    return U, V, W


def suN_generator(n):
    """
    Generate sorted SU(N) generators.
    
    Args:
        n (float): 
            Dimension of the system. 
        
    Returns:
        (list): 
            List of SU(N) generators.
    """
    symm, anti_symm, diag = suN_unsorted(n)
    if n == 2:
        return [symm[0], anti_symm[0], diag[0]]
    else:
        Lambda = [0.0] * len(symm + anti_symm + diag)

        Lambda[0], Lambda[1], Lambda[2] = symm[0], anti_symm[0], diag[0]

        repeat_times = 2
        m1, n1, k1 = 0, 3, 1
        while True:
            m1 += n1
            j, l = 0, 0
            for i in range(repeat_times):
                Lambda[m1 + j] = symm[k1]
                Lambda[m1 + j + 1] = anti_symm[k1]
                j += 2
                k1 += 1

            repeat_times += 1
            n1 = n1 + 2
            if k1 == len(symm):
                break

        m2, n2, k2 = 2, 5, 1
        while True:
            m2 += n2
            Lambda[m2] = diag[k2]
            n2 = n2 + 2
            k2 = k2 + 1
            if k2 == len(diag):
                break
        return Lambda


def gramschmidt(A):
    """
    Perform Gram-Schmidt orthogonalization.
    
    Args:
        A (list): 
            List of vectors to orthogonalize.
        
    Returns:
        (list): 
            List of orthonormal vectors.
    """
    dim = len(A)
    n = len(A[0])
    Q = [np.zeros(n, dtype=np.complex128) for i in range(dim)]
    for j in range(0, dim):
        q = A[j]
        for i in range(0, j):
            rij = np.vdot(Q[i], q)
            q = q - rij * Q[i]
        rjj = np.linalg.norm(q, ord=2)
        Q[j] = q / rjj
    return Q


def basis(dim, index):
    """
    Generate basis vector.
    
    Args:
        dim (float): 
            Dimension of Hilbert space.
        index (float): 
            Index of basis vector.
        
    Returns:
        (np.array): 
            Basis vector as column vector.
    """
    x = np.zeros(dim)
    x[index] = 1.0
    return x.reshape(dim, 1)


def sic_povm(fiducial):
    """
    Generate SIC-POVM from fiducial state.
    
    Args:
        fiducial (np.array): 
            the fiducial state vector.
        
    Returns:
        (list): 
            List of POVM elements.
    """
    d = fiducial.shape[0]
    w = np.exp(2.0 * np.pi * 1.0j / d)
    Z = np.diag([w**i for i in range(d)])
    X = np.zeros((d, d), dtype=np.complex128)
    for i in range(d):
        for j in range(d):
            if j != (d - 1):
                X += np.dot(basis(d, j + 1), basis(d, j).transpose().conj())
            else:
                X += np.dot(basis(d, 0), basis(d, j).transpose().conj())
    X = X / d

    D = [[[] for i in range(d)] for j in range(d)]
    for a in range(d):
        for b in range(d):
            X_a = np.linalg.matrix_power(X, a)
            Z_b = np.linalg.matrix_power(Z, b)
            phase = (-np.exp(1.0j * np.pi / d)) ** (a * b)
            D[a][b] = phase * np.dot(X_a, Z_b)

    res = []
    for m in range(d):
        for n in range(d):
            res_tp = np.dot(D[m][n], fiducial)
            res_tp = res_tp / np.linalg.norm(res_tp)
            basis_res = np.dot(res_tp, res_tp.conj().transpose()) / d
            res.append(basis_res)
    return res


def SIC(dim):
    """
    Generate SIC-POVM for given dimension.
    
    Args:
        dim (float): 
            Dimension of the system.
        
    Returns:
        (list): 
            List of SIC-POVM elements.
        
    Raises:
        ValueError: 
            If dimension > 151.
    """
    if dim <= 151:
        file_path = os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            "sic_fiducial_vectors/d%d.txt" % (dim),
        )
        data = np.loadtxt(file_path)
        fiducial = data[:, 0] + data[:, 1] * 1.0j
        fiducial = np.array(fiducial).reshape(len(fiducial), 1)
        M = sic_povm(fiducial)
        return M
    else:
        raise ValueError(
            "The dimension of the space should be less or equal to 151."
        )


def extract_ele(element, n):
    """
    Recursively extract elements.
    
    Args:
        element (np.array/list): 
            Input element
        n (int/float): 
            Depth of extraction.
    """
    if n:
        for x in element:
            yield from extract_ele(x, n - 1)
    else:
        yield element


def annihilation(n):
    """
    Create annihilation operator.
    
    Args:
        n (int/float): 
            Dimension of space.
        
    Returns:
        (np.array): 
            Annihilation operator matrix.
    """
    data = np.sqrt(np.arange(1, n, dtype=complex))
    indices = np.arange(1, n)
    indptr = np.arange(n + 1)
    indptr[-1] = n - 1
    a = csr_matrix((data, indices, indptr), shape=(n, n)).todense()
    return a


def brgd(n):
    """
    Generate binary reflected Gray code.
    
    Args:
        n (int/float): 
            Number of bits
        
    Returns:
        (list): 
            List of Gray code sequences
    """
    if n == 1:
        return ["0", "1"]
    L0 = brgd(n - 1)
    L1 = copy.deepcopy(L0)
    L1.reverse()
    L0 = ["0" + i for i in L0]
    L1 = ["1" + j for j in L1]
    res = L0 + L1
    return res


def BayesInput(x, func, dfunc, channel="dynamics"):
    """
    Generate input variables for Bayesian estimation.
    
    Args:
        x (np.array): 
            Parameter regimes
        func (callable): 
            Function returning H or K
        dfunc (callable): 
            Function returning dH or dK
        channel (str, optional): 
            "dynamics" or "Kraus" (default: "dynamics")
        
    Returns:
        (tuple): 
            Tuple of (H_list, dH_list) or (K_list, dK_list)
        
    Raises:
        ValueError: 
            For invalid channel.
    """
    x_all = product(*x)
    if channel == "dynamics":
        H_list, dH_list = [], []
        for xi in x_all:
            H_list.append(func(*xi))
            dH_list.append(dfunc(*xi))
        return H_list, dH_list
    elif channel == "Kraus":
        K_list, dK_list = [], []
        for xi in x_all:
            K_list.append(func(*xi))
            dK_list.append(dfunc(*xi))
        return K_list, dK_list
    else:
        raise ValueError(
            "{!r} is not a valid channel. Supported values: "
            "'dynamics' or 'Kraus'.".format(channel)
        )



import numpy as np
from more_itertools import zip_broadcast


def SpinSqueezing(rho, basis="Dicke", output="KU"):
    r"""
    Calculation of the spin squeezing parameter for a density matrix.

    The spin squeezing parameter $\xi$ is defined as:
    \begin{align}
    \xi^2 = \frac{N(\Delta J_{\vec{n}_1})^2}{\langle J_{\vec{n}_3}\rangle^2}
    \end{align}
    where $J_{\vec{n}_i}$ are the collective spin operators.

    Args:
        rho (array): 
            Density matrix.
        basis (str, optional): 
            Basis to use: "Dicke" (default) or "Pauli".
        output (str, optional): 
            Type of spin squeezing to calculate:  
                - "KU": Kitagawa-Ueda squeezing parameter.  
                - "WBIMH": Wineland et al. squeezing parameter.  

    Returns:
        (float): 
            Spin squeezing parameter.

    Raises:
        ValueError: If basis has invalid value.  
        NameError: If output has invalid value.  
    """
    N = len(rho) - 1
    coef = 4.0 / float(N)
    j = N / 2
    
    if basis == "Pauli":
        sp = np.array([[0.0, 1.0], [0.0, 0.0]])
        jp = []
        for i in range(N):
            if i == 0:
                jp_tp = np.kron(sp, np.identity(2 ** (N - 1)))
            elif i == N - 1:
                jp_tp = np.kron(np.identity(2 ** (N - 1)), sp)
            else:
                jp_tp = np.kron(
                    np.identity(2 ** i), 
                    np.kron(sp, np.identity(2 ** (N - 1 - i)))
                )
            jp.append(jp_tp)
        Jp = sum(jp)
    elif basis == "Dicke":
        offdiag = [
            np.sqrt(float(j * (j + 1) - m * (m + 1))) 
            for m in np.arange(j, -j - 1, -1)
        ][1:]
        # Ensure we create a complex array
        Jp = np.diag(offdiag, 1).astype(complex)
    else:
        valid_types = ["Dicke", "Pauli"]
        raise ValueError(
                f"Invalid basis: '{basis}'. Supported types: {', '.join(valid_types)}"
            )    
    
    Jx = 0.5 * (Jp + np.conj(Jp).T)
    Jy = -0.5 * 1j * (Jp - np.conj(Jp).T)
    Jz = np.diag(np.arange(j, -j - 1, -1))
    
    Jx_mean = np.trace(rho @ Jx)
    Jy_mean = np.trace(rho @ Jy)
    Jz_mean = np.trace(rho @ Jz)

    costheta = Jz_mean / np.sqrt(Jx_mean**2 + Jy_mean**2 + Jz_mean**2)
    sintheta = np.sin(np.arccos(costheta))
    cosphi = Jx_mean / np.sqrt(Jx_mean**2 + Jy_mean**2)
    sinphi = (np.sin(np.arccos(cosphi)) if Jy_mean > 0 
              else np.sin(2 * np.pi - np.arccos(cosphi)))
    
    Jn1 = -Jx * sinphi + Jy * cosphi
    Jn2 = (-Jx * costheta * cosphi 
           - Jy * costheta * sinphi 
           + Jz * sintheta)
    
    A = np.trace(rho @ (Jn1 @ Jn1 - Jn2 @ Jn2))
    B = np.trace(rho @ (Jn1 @ Jn2 + Jn2 @ Jn1))
    C = np.trace(rho @ (Jn1 @ Jn1 + Jn2 @ Jn2))
    
    V_minus = 0.5 * (C - np.sqrt(A**2 + B**2))
    V_minus = np.real(V_minus)
    xi = coef * V_minus
    xi = min(xi, 1.0)  # Cap at 1.0

    if output == "KU":
        pass
    elif output == "WBIMH":
        xi = (N / 2)**2 * xi / (Jx_mean**2 + Jy_mean**2 + Jz_mean**2)
    else:
        raise NameError("output should be either 'KU' or 'WBIMH'")

    return xi


def TargetTime(f, tspan, func, *args, **kwargs):
    r"""
    Calculation of the time to reach a given precision limit.

    This function finds the earliest time $t$ in `tspan` where the objective 
    function `func` reaches or crosses the target value $f$.

    Args:
        f (float): 
            The target value of the objective function.
        tspan (array): 
            Time points for the evolution.
        func (callable): 
            The objective function to evaluate. Must return a float.
        *args (tuple): 
            Positional arguments to pass to `func`.
        **kwargs (dict): 
            Keyword arguments to pass to `func`.

    Returns:
        (float): 
            Time to reach the given target precision.
    """

    args = list(zip_broadcast(*args))

    f_last = func(*(args[0]), **kwargs)
    idx = 1
    f_now = func(*(args[1]), **kwargs)

    while (f_now - f) * (f_last - f) > 0 and idx < (len(tspan) - 1):
        f_last = f_now
        idx += 1
        f_now = func(*(args[idx]), **kwargs)

    return tspan[idx]



import numpy as np
import warnings
import math
from quanestimation import QJL


class Lindblad:
    r"""
    Class for simulating quantum dynamics governed by the Lindblad master equation.

    The dynamics of a density matrix is described by the Lindblad master equation:
    
    \begin{aligned}
        \partial_t\rho &=\mathcal{L}\rho \nonumber \\
        &=-i[H,\rho]+\sum_i \gamma_i(\Gamma_i\rho\Gamma^{\dagger}_i-\frac{1}{2}
        \{\rho,\Gamma^{\dagger}_i \Gamma_i \}),
    \end{aligned}
    
    Symbols: 
        - $\rho$: the evolved density matrix
        - $H$: the Hamiltonian of the system
        - $\Gamma_i$: the $i$th decay operator
        - $\gamma_i$: the $i$th decay rate

    Attributes:
        tspan (np.array): 
            Time points for the evolution.
        rho0 (np.array): 
            Initial state (density matrix).
        H0 (np.array/list): 
            Free Hamiltonian. It is a matrix when time-independent, or a list of matrices 
            (with length equal to `tspan`) when time-dependent.
        dH (list): 
            Derivatives of the free Hamiltonian with respect to the unknown parameters.  
            Each element is a matrix representing the partial derivative with respect to 
            one parameter. For example, `dH[0]` is the derivative with respect to the 
            first parameter.
        decay (list): 
            Decay operators and corresponding decay rates. Input format:  
            `decay=[[Γ₁, γ₁], [Γ₂, γ₂], ...]`, where Γ₁, Γ₂ are decay operators and γ₁, γ₂ 
            are the corresponding decay rates.
        H (list): 
            Control Hamiltonians.
        ctrl (list, optional): 
            Control coefficients for each control Hamiltonian.
    """

    def __init__(self, tspan, rho0, H0, dH, decay=[], Hc=[], ctrl=[]):
        
        self.tspan = tspan
        self.rho0 = np.array(rho0, dtype=np.complex128)

        if type(H0) == np.ndarray:
            self.freeHamiltonian = np.array(H0, dtype=np.complex128)
        else:
            self.freeHamiltonian = [np.array(x, dtype=np.complex128) for x in H0]

        if type(dH[0]) != np.ndarray:
            raise TypeError("The derivative of Hamiltonian should be a list!")

        if not dH:
            dH = [np.zeros((len(self.rho0), len(self.rho0)))]
        self.Hamiltonian_derivative = [np.array(x, dtype=np.complex128) for x in dH]

        if not decay:
            decay_opt = [np.zeros((len(self.rho0), len(self.rho0)))]
            self.gamma = [0.0]
        else:
            decay_opt = [decay[i][0] for i in range(len(decay))]
            self.gamma = [decay[i][1] for i in range(len(decay))]
        self.decay_opt = [np.array(x, dtype=np.complex128) for x in decay_opt]

        if not Hc:
            Hc = [np.zeros((len(self.rho0), len(self.rho0)))]
            ctrl = [np.zeros(len(self.tspan) - 1)]
            self.control_Hamiltonian = [np.array(x, dtype=np.complex128) for x in Hc]
            self.control_coefficients = ctrl
        elif not ctrl:
            ctrl = [np.zeros(len(self.tspan) - 1) for j in range(len(Hc))]
            self.control_Hamiltonian = Hc
            self.control_coefficients = ctrl
        else:
            ctrl_length = len(ctrl)
            ctrlnum = len(Hc)
            if ctrlnum < ctrl_length:
                raise TypeError(
                    "There are %d control Hamiltonians but %d coefficients sequences: \
                                too many coefficients sequences"
                    % (ctrlnum, ctrl_length)
                )
            elif ctrlnum > ctrl_length:
                warnings.warn(
                    "Not enough coefficients sequences: there are %d control Hamiltonians \
                            but %d coefficients sequences. The rest of the control sequences are\
                            set to be 0."
                    % (ctrlnum, ctrl_length),
                    DeprecationWarning,
                )

            number = math.ceil((len(self.tspan) - 1) / len(ctrl[0]))
            if len(self.tspan) - 1 % len(ctrl[0]) != 0:
                tnum = number * len(ctrl[0])
                self.tspan = np.linspace(self.tspan[0], self.tspan[-1], tnum + 1)
            self.control_Hamiltonian = Hc
            self.control_coefficients = ctrl

    def expm(self):
        r"""
        Calculate the density matrix and its derivatives using the matrix exponential method.

        The density matrix at the $j$th time interval is obtained by:

        $$
            \rho_j = e^{\Delta t \mathcal{L}} \rho_{j-1}
        $$
        
        where $\Delta t$ is the time interval and $\rho_{j-1}$ is the density matrix 
        at the previous time step.

        The derivative $\partial_{\textbf{x}}\rho_j$ is calculated as:

        $$
            \partial_{\textbf{x}}\rho_j = \Delta t (\partial_{\textbf{x}}\mathcal{L}) \rho_j
            + e^{\Delta t \mathcal{L}} (\partial_{\textbf{x}}\rho_{j-1})
        $$

        Returns: 
            (tuple):
                rho (list): 
                    Density matrices at each time point in `tspan`.

                drho (list): 
                    Derivatives of the density matrices with respect to the unknown parameters.  
                    `drho[i][j]` is the derivative of the density matrix at the i-th time point 
                    with respect to the j-th parameter.
        """

        rho, drho = QJL.expm_py(
            self.tspan,
            self.rho0,
            self.freeHamiltonian,
            self.Hamiltonian_derivative,
            self.decay_opt,
            self.gamma,
            self.control_Hamiltonian,
            self.control_coefficients,
        )
        rho = [np.array(rho_i) for rho_i in rho]
        drho = [[np.array(drho_ij) for drho_ij in drho_i] for drho_i in drho]
        
        return rho, drho

    def ode(self):
        r"""
        Calculate the density matrix and its derivatives using an ODE solver.

        The density matrix at the $j$th time interval is obtained by:

        $$
            \rho_j = e^{\Delta t \mathcal{L}} \rho_{j-1},
        $$
        
        where $\Delta t$ is the time interval and $\rho_{j-1}$ is the density matrix 
        at the previous time step.

        The derivative $\partial_{\textbf{x}}\rho_j$ is calculated as:

        $$
            \partial_{\textbf{x}}\rho_j = \Delta t (\partial_{\textbf{x}}\mathcal{L}) \rho_j
            + e^{\Delta t \mathcal{L}} (\partial_{\textbf{x}}\rho_{j-1}).
        $$

        Returns:
            (tuple):
                rho (list): 
                    Density matrices at each time point in `tspan`.

                drho (list): 
                    Derivatives of the density matrices with respect to the unknown parameters.  
                    `drho[i][j]` is the derivative of the density matrix at the i-th time point 
                    with respect to the j-th parameter.
        """

        rho, drho = QJL.ode_py(
            self.tspan,
            self.rho0,
            self.freeHamiltonian,
            self.Hamiltonian_derivative,
            self.decay_opt,
            self.gamma,
            self.control_Hamiltonian,
            self.control_coefficients,
        )
        rho = [np.array(rho_i) for rho_i in rho]
        drho = [[np.array(drho_ij) for drho_ij in drho_i] for drho_i in drho]
        
        return rho, drho
        
    def secondorder_derivative(self, d2H):
        r"""
        Calculate the density matrix, its first derivatives, and second derivatives 
        with respect to the unknown parameters.

        The density matrix at the $j$th time interval is obtained by:

        $$
            \rho_j = e^{\Delta t \mathcal{L}} \rho_{j-1}.
        $$

        The first derivative $\partial_{\textbf{x}}\rho_j$ is calculated as:
       
        $$
            \partial_{\textbf{x}}\rho_j = \Delta t (\partial_{\textbf{x}}\mathcal{L}) \rho_j
            + e^{\Delta t \mathcal{L}} (\partial_{\textbf{x}}\rho_{j-1}).
        $$

        The second derivative $\partial_{\textbf{x}}^2\rho_j$ is calculated as:
        
        \begin{aligned}
            \partial_{\textbf{x}}^2\rho_j =& \Delta t (\partial_{\textbf{x}}^2\mathcal{L}) \rho_j \\
            &+ \Delta t (\partial_{\textbf{x}}\mathcal{L}) \partial_{\textbf{x}}\rho_j \\
            &+ \Delta t (\partial_{\textbf{x}}\mathcal{L}) e^{\Delta t \mathcal{L}} \partial_{\textbf{x}}\rho_{j-1} \\
            &+ e^{\Delta t \mathcal{L}} (\partial_{\textbf{x}}^2\rho_{j-1})
        \end{aligned}

        Args:
            d2H (list): Second-order derivatives of the free Hamiltonian with respect to the unknown parameters.  
                Each element is a matrix representing the second partial derivative with respect to 
                two parameters. For example, `d2H[0]` might be the second derivative with respect to 
                the first parameter twice, or a mixed partial derivative.

        Returns:
            (tuple):
                rho (list): 
                    Density matrices at each time point in `tspan`.

                drho (list): 
                    First derivatives of the density matrices with respect to the unknown parameters.
                    
                d2rho (list): 
                    Second derivatives of the density matrices with respect to the unknown parameters.
        """

        d2H = [np.array(x, dtype=np.complex128) for x in d2H]
        rho, drho, d2rho = QJL.secondorder_derivative(
            self.tspan,
            self.rho0,
            self.freeHamiltonian,
            self.Hamiltonian_derivative,
            d2H,
            self.decay_opt,
            self.gamma,
            self.control_Hamiltonian,
            self.control_coefficients,
        )
        rho = [np.array(rho_i) for rho_i in rho]
        drho = [[np.array(drho_ij) for drho_ij in drho_i] for drho_i in drho]
        
        return rho, drho, d2rho



from quanestimation import QJL
import quanestimation.MeasurementOpt.MeasurementStruct as Measurement


class AD_Mopt(Measurement.MeasurementSystem):
    """
    Attributes
    ----------
    > **savefile:** `bool`
        -- Whether or not to save all the measurements.  
        If set `True` then the measurements and the values of the 
        objective function obtained in all episodes will be saved during 
        the training. If set `False` the measurement in the final 
        episode and the values of the objective function in all episodes 
        will be saved.

    > **Adam:** `bool`
        -- Whether or not to use Adam for updating measurements.

    > **measurement0:** `list of arrays`
        -- Initial guesses of measurements.

    > **max_episode:** `int`
        -- The number of episodes.
  
    > **epsilon:** `float`
        -- Learning rate.

    > **beta1:** `float`
        -- The exponential decay rate for the first moment estimates.

    > **beta2:** `float`
        -- The exponential decay rate for the second moment estimates.

    > **eps:** `float`
        -- Machine epsilon.

    > **load:** `bool`
        -- Whether or not to load measurements in the current location.  
        If set `True` then the program will load measurement from "measurements.csv"
        file in the current location and use it as the initial measurement.
    """

    def __init__(
        self,
        mtype,
        minput,
        savefile=False,
        Adam=False,
        measurement0=[],
        max_episode=300,
        epsilon=0.01,
        beta1=0.90,
        beta2=0.99,
        seed=1234,
        eps=1e-8,
        load=False,
    ):

        Measurement.MeasurementSystem.__init__(
            self, mtype, minput, savefile, measurement0, seed, eps, load 
        )

        self.Adam = Adam
        self.max_episode = max_episode
        self.epsilon = epsilon
        self.beta1 = beta1
        self.beta2 = beta2
        self.mt = 0.0
        self.vt = 0.0
        self.seed = seed

        if self.Adam:
            self.alg = QJL.AD(
                self.max_episode, self.epsilon, self.beta1, self.beta2
            )
        else:
            self.alg = QJL.AD(self.max_episode, self.epsilon)

    def CFIM(self, W=[]):
        r"""
        Choose CFI or $\mathrm{Tr}(WI^{-1})$ as the objective function. 
        In single parameter estimation the objective function is CFI and 
        in multiparameter estimation it will be $\mathrm{Tr}(WI^{-1})$.

        Parameters
        ----------
        > **W:** `matrix`
            -- Weight matrix.
        """
        
        if self.mtype == "projection":
            raise ValueError(
                "AD is not available when mtype is projection. Supported methods are 'PSO' and 'DE'.",
            )
        else:
            super().CFIM(W)



import pytest
import numpy as np
from quanestimation.AsymptoticBound.CramerRao import (
    QFIM, CFIM, QFIM_Kraus, QFIM_Bloch, QFIM_Gauss, LLD, RLD, FIM, FI_Expt, SLD
)


def test_CramerRao_SLD():
    """
    Test Cramer-Rao bound for parameterized quantum state.
    Checks QFIM and CFIM calculation for specific state and derivatives.
    """
    # Parameterized state
    theta = np.pi / 4
    phi = np.pi / 4
    rho = np.array([
        [np.cos(theta)**2, np.cos(theta) * np.sin(theta) * np.exp(-1j * phi)],
        [np.cos(theta) * np.sin(theta) * np.exp(1j * phi), np.sin(theta)**2]
    ])
    
    # State derivatives
    drho = [
        np.array([
            [-np.sin(2 * theta), 2 * np.cos(2 * theta) * np.exp(-1j * phi)],
            [2 * np.cos(2 * theta) * np.exp(1j * phi), np.sin(2 * theta)]
        ]),
        np.array([
            [0, -1j * np.cos(theta) * np.sin(theta) * np.exp(-1j * phi)],
            [1j * np.cos(theta) * np.sin(theta) * np.exp(1j * phi), 0]
        ])
    ]
    
    # Calculate QFIM
    result = QFIM(rho, drho, LDtype="SLD")
    # Measurement operators
    M = [np.array([[1.0, 0.0], [0.0, 0.0]]), 
         np.array([[0.0, 0.0], [0.0, 1.0]])]
    # Calculate CFIM
    resultc = CFIM(rho, drho, M)
    
    # Verify results
    expected_qfim = np.array([[4.0, 0.0], [0.0, np.sin(2 * theta)**2]])
    expected_cfim = np.array([[4.0, 0.0], [0.0, 0.0]])
    assert np.allclose(result, expected_qfim)
    assert np.allclose(resultc, expected_cfim)

    # Test exportLD option
    result, SLD_res = QFIM(rho, drho, LDtype="SLD", exportLD=True)
    expected_SLD = SLD(rho, drho)
    assert np.allclose(SLD_res, expected_SLD)

    # Test invalid LDtype
    with pytest.raises(ValueError):
        QFIM(rho, drho, LDtype="invalid")
    with pytest.raises(ValueError):
        QFIM(rho, drho, LDtype="RLD")
    with pytest.raises(ValueError):
        QFIM(rho, drho, LDtype="LLD")


def test_CFIM_singleparameter():
    """
    Test Classical Fisher Information Matrix for single parameter.
    Checks CFIM calculation for specific state and derivative.
    """
    # Parameterized state
    theta = np.pi / 4
    rho = np.array([
        [np.cos(theta)**2, np.sin(theta) * np.cos(theta)],
        [np.sin(theta) * np.cos(theta), np.sin(theta)**2]
    ])
    
    # State derivative
    drho = [np.array([
        [-np.sin(2 * theta), 2 * np.cos(2 * theta)],
        [2 * np.cos(2 * theta), np.sin(2 * theta)]
    ])]
    
    # Calculate CFIM
    result = CFIM(rho, drho, [])
    assert np.allclose(result, 2.0)


def test_QFIM_Kraus():
    """
    Test QFIM for Kraus representation.
    Checks QFIM calculation for specific Kraus operators.
    """
    # Kraus operators
    K0 = np.array([[1, 0], [0, np.sqrt(0.5)]])
    K1 = np.array([[np.sqrt(0.5), 0], [0, 0]])
    K = [K0, K1]
    
    # Kraus operator derivatives
    dK = [
        [np.array([[0, 0], [0, -0.5 / np.sqrt(0.5)]])],
        [np.array([[0, 0.5 / np.sqrt(0.5)], [0, 0]])]
    ]
    
    # Probe state
    rho0 = 0.5 * np.array([[1.0, 1.0], [1.0, 1.0]])
    
    # Calculate QFIM
    result = QFIM_Kraus(rho0, K, dK)
    assert np.allclose(result, 1.5)


def test_QFIM_Bloch():
    """
    Test QFIM for Bloch vector representation.
    Checks QFIM calculation for specific Bloch vector and derivatives.
    """
    # Bloch vector parameters
    theta = np.pi / 4
    phi = np.pi / 2
    eta = 0.8
    b = eta * np.array([
        np.sin(2 * theta) * np.cos(phi),
        np.sin(2 * theta) * np.sin(phi),
        np.cos(2 * theta)
    ])
    
    # Bloch vector derivatives
    db_theta = eta * np.array([
        2 * np.cos(2 * theta) * np.cos(phi),
        2 * np.cos(2 * theta) * np.sin(phi),
        -2 * np.sin(2 * theta)
    ])
    db_phi = eta * np.array([
        -np.sin(2 * theta) * np.sin(phi),
        np.sin(2 * theta) * np.cos(phi),
        0
    ])
    db = [db_theta, db_phi]
    
    # Calculate QFIM
    result = QFIM_Bloch(b, db)
    expected = np.array([
        [4.0 * eta**2, 0.0],
        [0.0, eta**2 * np.sin(2 * theta)**2]
    ])
    assert np.allclose(result, expected)


def test_QFIM_Bloch_pure():
    """
    Test QFIM for pure state in Bloch representation.
    Checks QFIM calculation for specific Bloch vector and derivatives.
    """
    # Bloch vector parameters
    theta = np.pi / 4
    phi = np.pi / 2
    eta = 1.0
    b = eta * np.array([
        np.sin(2 * theta) * np.cos(phi),
        np.sin(2 * theta) * np.sin(phi),
        np.cos(2 * theta)
    ])
    
    # Bloch vector derivatives
    db_theta = eta * np.array([
        2 * np.cos(2 * theta) * np.cos(phi),
        2 * np.cos(2 * theta) * np.sin(phi),
        -2 * np.sin(2 * theta)
    ])
    db_phi = eta * np.array([
        -np.sin(2 * theta) * np.sin(phi),
        np.sin(2 * theta) * np.cos(phi),
        0
    ])
    db = [db_theta, db_phi]
    
    # Calculate QFIM
    result = QFIM_Bloch(b, db)
    expected = np.array([
        [4.0 * eta**2, 0.0],
        [0.0, eta**2 * np.sin(2 * theta)**2]
    ])
    assert np.allclose(result, expected)


def test_QFIM_Bloch_highdimension():
    """
    Test QFIM for high-dimensional Bloch vector.
    Checks QFIM calculation for specific Bloch vector and derivative.
    """
    # Bloch vector parameters
    theta = np.pi / 4
    phi = np.pi / 2
    b = np.array([
        np.sin(2 * theta) * np.cos(phi),
        np.sin(2 * theta) * np.sin(phi),
        np.cos(2 * theta),
        0.0, 0.0, 0.0, 0.0, 0.0
    ])
    
    # Bloch vector derivative
    db = [np.array([
        2 * np.cos(2 * theta) * np.cos(phi),
        2 * np.cos(2 * theta) * np.sin(phi),
        -2 * np.sin(2 * theta),
        0.0, 0.0, 0.0, 0.0, 0.0
    ])]
    
    # Calculate QFIM
    result = QFIM_Bloch(b, db)
    assert np.allclose(result, 8.0)

    b_invalid = np.array([
        0.0, 0.0, 0.0, 0.0, 0.0
    ])
    with pytest.raises(ValueError):
        QFIM_Bloch(b_invalid, db)


def test_QFIM_Gauss_multiparameter():
    """
    Test QFIM for Gaussian state with multiple parameters.
    Checks QFIM calculation for specific Gaussian state.
    """
    # Gaussian state parameters
    r = 0.8  # squeezing parameter
    beta = 0.5
    lamb = 1 / np.tanh(beta / 2)
    mu = np.array([0.0, 0.0])
    sigma = lamb * np.array([
        [np.cosh(2 * r), -np.sinh(2 * r)],
        [-np.sinh(2 * r), np.cosh(2 * r)]
    ])
    
    # Derivatives
    dmu = [np.array([0.0, 0.0]), np.array([0.0, 0.0])]
    dlamb = -0.5 / (np.sinh(beta / 2)**2)
    dsigma = [
        dlamb * np.array([
            [np.cosh(2 * r), -np.sinh(2 * r)],
            [-np.sinh(2 * r), np.cosh(2 * r)]
        ]),
        lamb * 2 * np.array([
            [np.sinh(2 * r), -np.cosh(2 * r)],
            [-np.cosh(2 * r), np.sinh(2 * r)]
        ])
    ]
    
    # Calculate QFIM
    result = QFIM_Gauss(mu, dmu, sigma, dsigma)
    expected = np.array([
        [(lamb**2 - 1)**2 / (2 * (4 * lamb**2 - 1)), 0.0],
        [0.0, 8 * lamb**2 / (4 * lamb**2 + 1)]
    ])
    assert np.allclose(result, expected)


def test_QFIM_Gauss_singleparameter():
    """
    Test QFIM for Gaussian state with single parameter.
    Checks QFIM calculation for specific Gaussian state.
    """
    # Gaussian state parameters
    r = 0.8  # squeezing parameter
    beta = 0.5
    lamb = 1 / np.tanh(beta / 2)
    mu = np.array([0.0, 0.0])
    sigma = lamb * np.array([
        [np.cosh(2 * r), -np.sinh(2 * r)],
        [-np.sinh(2 * r), np.cosh(2 * r)]
    ])
    
    # Derivatives
    dmu = [np.array([0.0, 0.0])]
    dlamb = -0.5 / (np.sinh(beta / 2)**2)
    dsigma = [dlamb * np.array([
        [np.cosh(2 * r), -np.sinh(2 * r)],
        [-np.sinh(2 * r), np.cosh(2 * r)]
    ])]
    
    # Calculate QFIM
    result = QFIM_Gauss(mu, dmu, sigma, dsigma)
    expected = (lamb**2 - 1)**2 / (2 * (4 * lamb**2 - 1))
    assert np.allclose(result, expected)


def test_QFIM_LLD_singleparameter():
    """
    Test left logarithmic derivative for specific state.
    Checks LLD calculation and QFIM result.
    """
    # State parameters
    theta = np.pi / 4
    phi = np.pi / 4
    eta = 0.8
    rho = 0.5 * np.array([
        [1 + eta * np.cos(2 * theta), 
         eta * np.sin(2 * theta) * np.exp(-1j * phi)],
        [eta * np.sin(2 * theta) * np.exp(1j * phi), 
         1 - eta * np.cos(2 * theta)]
    ])
    
    # State derivative
    drho = [0.5 * np.array([
        [0.0, -1j * eta * np.sin(2 * theta) * np.exp(-1j * phi)],
        [1j * eta * np.sin(2 * theta) * np.exp(1j * phi), 0.0]
    ])]
    
    # Calculate LLD
    result = LLD(rho, drho, rep="original")
    expected = (1 / (1 - eta**2)) * np.array([
        [1j * eta**2 * np.sin(2 * theta)**2, 
         -1j * eta * (1 + eta * np.cos(2 * theta)) * np.sin(2 * theta) * np.exp(-1j * phi)],
        [1j * eta * (1 - eta * np.cos(2 * theta)) * np.sin(2 * theta) * np.exp(1j * phi), 
         -1j * eta**2 * np.sin(2 * theta)**2]
    ])
    assert np.allclose(result, expected)

    # Calculate QFIM
    result_QFIM = QFIM(rho, drho, LDtype="LLD")
    expected_QFIM = eta**2 * np.sin(2 * theta)**2 * (3 * eta**2 + 1) / (1 - eta**2)**2
    assert np.allclose(result_QFIM, expected_QFIM)

    # Test eigen representation
    result_eigen = LLD(rho, drho, rep="eigen")
    _, vec = np.linalg.eig(rho)
    expected_eigen = vec.conj().transpose() @ result @ vec
    assert np.allclose(result_eigen, expected_eigen)

    # Test invalid representation
    with pytest.raises(ValueError):
        LLD(rho, drho, rep="invalid")


def test_QFIM_RLD_singleparameter():
    """
    Test right logarithmic derivative for specific state.
    Checks RLD calculation and QFIM result.
    """
    # State parameters
    theta = np.pi / 4
    phi = np.pi / 4
    eta = 0.8
    rho = 0.5 * np.array([
        [1 + eta * np.cos(2 * theta), 
         eta * np.sin(2 * theta) * np.exp(-1j * phi)],
        [eta * np.sin(2 * theta) * np.exp(1j * phi), 
         1 - eta * np.cos(2 * theta)]
    ])
    
    # State derivative
    drho = [0.5 * np.array([
        [0.0, -1j * eta * np.sin(2 * theta) * np.exp(-1j * phi)],
        [1j * eta * np.sin(2 * theta) * np.exp(1j * phi), 0.0]
    ])]
    
    # Calculate RLD
    result = RLD(rho, drho, rep="original")
    expected = (1 / (1 - eta**2)) * np.array([
        [-1j * eta**2 * np.sin(2 * theta)**2, 
         -1j * eta * (1 - eta * np.cos(2 * theta)) * np.exp(-1j * phi) * np.sin(2 * theta)],
        [1j * eta * (1 + np.cos(2 * theta)) * np.exp(1j * phi) * np.sin(2 * theta), 
         1j * eta**2 * np.sin(2 * theta)**2]
    ])
    assert np.allclose(result, expected)

    # Calculate QFIM
    result_QFIM = QFIM(rho, drho, LDtype="RLD")
    expected_QFIM = eta**2 * np.sin(2 * theta)**2 / (1.0 - eta**2)
    assert np.allclose(result_QFIM, expected_QFIM)

    # Test eigen representation
    result_eigen = RLD(rho, drho, rep="eigen")
    _, vec = np.linalg.eig(rho)
    expected_eigen = vec.conj().transpose() @ result @ vec
    assert np.allclose(result_eigen, expected_eigen)

    # Test invalid representation
    with pytest.raises(ValueError):
        RLD(rho, drho, rep="invalid")


def test_FIM_singleparameter():
    """
    Test Fisher Information Matrix (FIM) for single parameter.
    Checks FIM calculation for classical probability distribution.
    """
    x = 1.0
    theta = np.pi / 3
    p = np.array([
        np.cos(x * theta)**2, 
        np.sin(x * theta)**2
    ])
    dp = [np.array([
        -x * np.sin(2 * x * theta), 
        x * np.sin(2 * x * theta)
    ])]
    result = FIM(p, dp)
    assert np.allclose(result, 4.0)

def test_FIM_multiparameter():
    """
    Test Fisher Information Matrix (FIM) for multiple parameters.
    Checks FIM calculation for classical probability distribution.
    """
    x = 1.0
    theta = np.pi / 3
    p = np.array([
        np.cos(x * theta)**2, 
        np.sin(x * theta)**2
    ])
    dp = [
        np.array([
            -theta * np.sin(2 * x * theta), 
            theta * np.sin(2 * x * theta)
        ]),
        np.array([
            -x * np.sin(2 * x * theta), 
            x * np.sin(2 * x * theta)
        ])
    ]
    result = FIM(p, dp)
    expected = np.array([
        [4.38649084, 4.1887902],
        [4.1887902, 4.0]
    ])
    assert np.allclose(result, expected)

def test_FI_Expt():
    """
    Test Fisher Information calculation for experimental data.
    Checks FI_Expt with different distribution types.
    """
    # Normal, Gamma, and Rayleigh distributions
    dx = 0.001
    y1_norm = np.random.normal(loc=0.0, scale=1.0, size=1000)
    y2_norm = np.random.normal(loc=dx, scale=1.0, size=1000)
    
    result_norm = FI_Expt(y1_norm, y2_norm, dx, ftype="norm")
    result_gamma = FI_Expt(y1_norm, y2_norm, dx, ftype="gamma")
    result_rayleigh = FI_Expt(y1_norm, y2_norm, dx, ftype="rayleigh")

    # Poisson distribution
    dx_poisson = 0.001
    y1_poi = np.random.poisson(lam=1, size=1000)
    y2_poi = np.random.poisson(lam=1 + dx_poisson, size=1000)
    result_poisson = FI_Expt(y1_poi, y2_poi, dx_poisson, ftype="poisson")

    # Verify results are floats
    assert all(isinstance(res, float) for res in [
        result_norm, result_gamma, result_rayleigh, result_poisson
    ])

    # Test invalid distribution type
    with pytest.raises(ValueError):
        FI_Expt(y1_norm, y2_norm, dx, ftype="invalid")

def test_invalid_input():
    """
    Test input validation for functions in the Cramer-Rao module.
    Verifies appropriate errors are raised for invalid inputs.
    """
    # Test CFIM with invalid inputs
    with pytest.raises(TypeError):
        CFIM(
            np.array([[1, 0], [0, 1]]),
            np.array([[1, 0], [0, 1]]),
            None
        )

    # Test QFIM with invalid LDtype
    with pytest.raises(ValueError):
        QFIM(
            np.array([[1, 0], [0, 1]]),
            [np.array([[1, 0], [0, 1]])],
            LDtype="invalid"
        )

    # Test QFIM with invalid drho
    with pytest.raises(TypeError):
        QFIM(np.array([[1, 0], [0, 1]]), None)

    # Test QFIM_Bloch with invalid db
    with pytest.raises(TypeError):
        QFIM_Bloch(np.array([[1.0], [0.0], [1.0]]), None)

    # Test logarithmic derivatives with invalid drho
    with pytest.raises(TypeError):
        SLD(np.array([[1, 0], [0, 1]]), None)

    with pytest.raises(TypeError):
        LLD(np.array([[1, 0], [0, 1]]), None)

    with pytest.raises(TypeError):
        RLD(np.array([[1, 0], [0, 1]]), None)



from quanestimation.ControlOpt.ControlStruct import (
    ControlSystem,
    ControlOpt,
    csv2npy_controls,
)
from quanestimation.ControlOpt.GRAPE_Copt import (
    GRAPE_Copt,
)
from quanestimation.ControlOpt.DE_Copt import (
    DE_Copt,
)
from quanestimation.ControlOpt.PSO_Copt import (
    PSO_Copt,
)
# from quanestimation.ControlOpt.DDPG_Copt import (
#     DDPG_Copt,
# )

__all__ = [
    "ControlSystem",
    "ControlOpt",
    "GRAPE_Copt",
    "DE_Copt",
    "PSO_Copt",
    # "DDPG_Copt",
    "csv2npy_controls",
]



import numpy as np


def Kraus(rho0, K, dK):
    r"""
    Parameterization of a quantum state using Kraus operators.

    The evolved density matrix $\rho$ is given by
    
    \begin{aligned}
        \rho=\sum_i K_i \rho_0 K_i^{\dagger},
    \end{aligned}

    where $\rho_0$ is the initial density matrix and $K_i$ are the Kraus operators.

    Args: 
        rho0 (np.array): 
            Initial density matrix.
        K (list): 
            Kraus operators.
        dK (list): 
            Derivatives of the Kraus operators with respect to the unknown parameters to be 
            estimated. This is a nested list where the first index corresponds to the Kraus operator 
            and the second index corresponds to the parameter. For example, `dK[0][1]` is the derivative 
            of the second Kraus operator with respect to the first parameter.

    Returns:
        (tuple):
            rho (np.array): 
                Evolved density matrix.
                
            drho (list): 
                Derivatives of the evolved density matrix with respect to the unknown parameters.  
                Each element in the list is a matrix representing the partial derivative of $\rho$ with 
                respect to one parameter.
    """

    k_num = len(K)
    para_num = len(dK[0])
    dK_reshape = [[dK[i][j] for i in range(k_num)] for j in range(para_num)]

    rho = sum([np.dot(Ki, np.dot(rho0, Ki.conj().T)) for Ki in K])
    drho = [sum([(np.dot(dKi, np.dot(rho0, Ki.conj().T))+ np.dot(Ki, np.dot(rho0, dKi.conj().T))) for (Ki, dKi) in zip(K, dKj)]) for dKj in dK_reshape]

    return rho, drho



import numpy as np
from quanestimation import QJL
from quanestimation.Common.Common import brgd, annihilation


class Adapt_MZI:
    """
    Attributes
    ----------
    > **x:** `list`
        -- The regimes of the parameters for the integral.

    > **p:** `multidimensional array`
        -- The prior distribution.

    > **rho0:** `matrix`
        -- Initial state (density matrix).

    """
    def __init__(self, x, p, rho0):

        self.x = x
        self.p = p
        self.rho0 = rho0
        self.N = int(np.sqrt(len(rho0))) - 1
        # self.a = annihilation(self.N + 1)

    def general(self):
        self.MZI_type = "general"

    def online(self, target="sharpness", output="phi"):
        """
        Parameters
        ----------
        > **target:** `string`
            -- Setting the target function for calculating the tunable phase. Options are:  
            "sharpness" (default) -- Sharpness.  
            "MI" -- Mutual information. 

        > **output:** `string`
            -- The output the class. Options are:  
            "phi" (default) -- The tunable phase.  
            "dphi" -- Phase difference. 
        """
        phi = QJL.adaptMZI_online(
            self.x, self.p, self.rho0, output, target
        )

    def offline(
        self,
        target="sharpness",
        method="DE",
        p_num=10,
        deltaphi0=[],
        c=1.0,
        cr=0.5,
        c0=1.0,
        c1=2.0,
        c2=2.0,
        seed=1234,
        max_episode=1000,
        eps=1e-8,
    ):
        """
        Parameters
        ----------
        > **target:** `string`
            -- Setting the target function for calculating the tunable phase. Options are:  
            "sharpness" (default) -- Sharpness.  
            "MI" -- Mutual information. 

        > **method:** `string`
            -- The method for the adaptive phase estimation. Options are:  
            "DE" (default) -- DE algorithm for the adaptive phase estimation.    
            "PSO" -- PSO algorithm for the adaptive phase estimation.

        If the `method=DE`, the parameters are:
        > **p_num:** `int`
            -- The number of populations.

        > **deltaphi0:** `list`
            -- Initial guesses of phase difference.

        > **max_episode:** `int`
            -- The number of episodes.
  
        > **c:** `float`
            -- Mutation constant.

        > **cr:** `float`
            -- Crossover constant.

        > **seed:** `int`
            -- Random seed.

        > **eps:** `float`
            -- Machine epsilon.
        
        If the `method=PSO`, the parameters are:

        > **deltaphi0:** `list`
            -- Initial guesses of phase difference.

        > **max_episode:** `int or list`
            -- If it is an integer, for example max_episode=1000, it means the 
            program will continuously run 1000 episodes. However, if it is an
            array, for example max_episode=[1000,100], the program will run 
            1000 episodes in total but replace states of all  the particles 
            with global best every 100 episodes.
  
        > **c0:** `float`
            -- The damping factor that assists convergence, also known as inertia weight.

        > **c1:** `float`
            -- The exploitation weight that attracts the particle to its best previous 
            position, also known as cognitive learning factor.

        > **c2:** `float`
            -- The exploitation weight that attracts the particle to the best position  
            in the neighborhood, also known as social learning factor.

        > **eps:** `float`
            -- Machine epsilon.
        """
        comb_tp = brgd(self.N)
        comb = [
            np.array([int(list(comb_tp[i])[j]) for j in range(self.N)])
            for i in range(2**self.N)
        ]
        
        if method == "DE":
            QJL.DE_deltaphiOpt(
                self.x,
                self.p,
                self.rho0,
                comb,
                p_num,
                deltaphi0,
                c,
                cr,
                seed,
                max_episode,
                target,
                eps,
            )
        elif method == "PSO":
            QJL.PSO_deltaphiOpt(
                self.x,
                self.p,
                self.rho0,
                comb,
                p_num,
                deltaphi0,
                c0,
                c1,
                c2,
                seed,
                max_episode,
                target,
                eps,
            )
        else:
            raise ValueError(
                "{!r} is not a valid value for method, supported values are 'DE' and 'PSO'.".format(
                    method
                )
            )
